#include <csrutils.h>
#include <csrstream.h>
#include <csrmacro.h>

using namespace csrutils;

BufferStream::BufferStream() {
	s_data    = NULL;
	s_bufsize = 0;
	s_size    = 0;
	s_offset  = 0;
}

BufferStream::~BufferStream() {
	if( s_data )
		delete[] s_data;
}

uint64_t BufferStream::getsize() {
	return s_size;
}

uint64_t BufferStream::getposition() {
	return s_offset;
}

bool BufferStream::setposition(uint64_t pos) {
	if( pos > (uint64_t)s_size )
		return false;
	s_offset = (size_t) pos;
	return true;
}

void *BufferStream::getdata() {
	return s_data;
}

bool BufferStream::setend() {
	s_size = s_offset;
	return true;
}

int BufferStream::read(void *data, size_t size, size_t& read) {
	size_t avail = MIN(s_size-s_offset,size);
	read = avail;
	memcpy(data,s_data+s_offset,avail);
	s_offset += avail;
	if( s_offset==s_size )
		return IO_EOF;
	else
		return IO_OK;
}

int BufferStream::write(const void *data, size_t size, size_t& written) {
	if( s_offset + size > s_bufsize ) {
		size_t new_size = s_offset + size;
		char *new_buf = new char[new_size];
		if( s_data!=NULL ) {
			memcpy(new_buf,s_data,s_size);
			delete[] s_data;
		}
		s_data = new_buf;
		s_bufsize = new_size;
	}
	if( s_offset + size > s_size )
		s_size = s_offset + size;
	memcpy(s_data+s_offset,data,size);
	s_offset += size;
	written = size;
	return IO_OK;
}
