#define _CRT_SECURE_NO_WARNINGS
#define _CRT_NON_CONFORMING_SWPRINTFS

#include <wchar.h>

#include "csrdomstore.h"
#include "csrmacro.h"

using namespace csrutils;

// TODO: support of items in DOMSWrite

struct FDOMS_OBJECT {
	uint64_t size;
	uint32_t nitems;
	uint32_t res0;
};

struct FDOMS_ITEM {
	uint64_t size;
	uint64_t subsize;
	uint64_t datasize;
	uint32_t szname;
	uint32_t type;
	uint32_t number;
	uint32_t res0;
};

DOMSObject::DOMSObject() {
}

DOMSObject::DOMSObject(const DOMSObject& obj) {
	for( auto iter=obj.params.begin(); iter!=obj.params.end(); iter++ ) {
		params[ (*iter).first ] = new DOMSItemHolder( *(*iter).second );
	}
}

DOMSObject::~DOMSObject() {
}

uint64_t DOMSObject::size() const {
	uint64_t size = sizeof(FDOMS_OBJECT);
	for( auto iter=items.begin(); iter!=items.end(); iter++ )
		size += (*iter)->size();
	for( auto iter=params.begin(); iter!=params.end(); iter++ )
		size += (*iter).second->size();
	return size;
}

uint64_t DOMSObject::store(void* data, uint64_t avail) const {
	uint64_t size = sizeof(FDOMS_OBJECT);
	if( avail < size )
		return 0;
	FDOMS_OBJECT* o = (FDOMS_OBJECT*) data;
	o->nitems = (uint32_t)( items.size() + params.size() );
	char* data0 = (char*) data;
	for( auto iter=items.begin(); iter!=items.end(); iter++ ) {
		uint64_t res = (*iter)->store(data0+size,avail-size);
		if( res==0 )
			return 0;
		size += res;
	}
	for( auto iter=params.begin(); iter!=params.end(); iter++ ) {
		uint64_t res = (*iter).second->store(data0+size,avail-size);
		if( res==0 )
			return 0;
		size += res;
	}
	o->size = size;
	o->res0 = 0;
	return size;
}

uint64_t DOMSObject::load(void* data, uint64_t avail) {
	uint64_t size = sizeof(FDOMS_OBJECT);
	if( avail < size )
		return 0;
	FDOMS_OBJECT* o = (FDOMS_OBJECT*) data;
	int nitems = (int) o->nitems;
	uint64_t rsize = o->size;
	char* data0 = (char*) data;
	for( int i=0; i<nitems; i++ ) {
		uint64_t res;
		DOMSItemHolder* h = DOMSItemHolder::load2(data0+size,rsize-size,res);
		if( res==0 ) {
			return 0;
		}
		size += res;
		if( h->use_name )
			params[ h->name ] = h;
		else {
			//items[ h->number ] = h;
			h->number = (int) items.size();
			items.push_back(h);
		}
	}
	if( rsize!=size)
		return 0;
	return rsize;
}

void DOMSObject::print(FILE* fd, int sublevel) {
	wchar_t* spaces = new wchar_t[sublevel+1];
	for( int i=0; i<sublevel; i++ )
		spaces[i] = ' ';
	//memset(spaces,' ',sublevel);
	spaces[sublevel] = '\0';
	fwprintf(fd,L"%s{\r\n",spaces);
	for( auto iter=items.begin(); iter!=items.end(); iter++ )
		(*iter)->print(fd,sublevel+1);
	for( auto iter=params.begin(); iter!=params.end(); iter++ )
		(*iter).second->print(fd,sublevel+1);
	fwprintf(fd,L"%s}\r\n",spaces);
	delete[] spaces;
}

//DOMSItem DOMSObject::operator[] (const wchar_t* name) {
//	return DOMSItem(*this,name);
//}

DOMSItem DOMSObject::operator[] (const String& name) {
	return DOMSItem(*this,name);
}

DOMSItem DOMSObject::operator[] (int id) {
	return DOMSItem(*this,id);
}

DOMSIterator::DOMSIterator(DOMSObject& _object)
		: object(_object), veciter(_object.items.begin()), vecbegin(_object.items.begin()), vecend(_object.items.end()), mapiter(_object.params.begin()), mapbegin(_object.params.begin()), mapend(_object.params.end()) {
}

void DOMSIterator::operator++ (int) {
	if( veciter != vecend )
		veciter ++;
	else if( mapiter != mapend )
			mapiter ++;
}

DOMSItem DOMSIterator::operator* () {
	if( veciter != vecend )
		return DOMSItem((DOMSObject&)object,(*veciter)->number);
	else
		return DOMSItem((DOMSObject&)object,(*mapiter).second->name);
}

DOMSIterator::operator bool() const {
	return (veciter != vecend) || (mapiter != mapend);
}

void DOMSIterator::reset() {
	mapiter = mapbegin;
}

bool DOMSIterator::has_next() const {
	return mapiter != mapend;
}

DOMSIterator DOMSObject::iter() {
	DOMSIterator iter(*this);
	return iter;
}

DOMSItemHolder& DOMSObject::provideParam(const String& name) {
	size_t ppos = name.strpos('.');
	if( ppos==-1 ) {
		if( params.find(name)!=params.end() )
			return *params[name];
		else {
			DOMSItemHolder* h = new DOMSItemHolder(name);
			params[ h->name.bytes() ] = h;
			return *params[name];
		}
	} else {
		StringB name1 = name.substr(0,ppos);
		StringB name2 = name.substr(ppos+1);
		DOMSItemHolder* h;
		if( params.find(name1)!=params.end() )
			h = params[name1];
		else {
			h = new DOMSItemHolder(name1);
			params[ h->name.bytes() ] = h;
			h->set( DOMSObject() );
		}
		if( h->type != DOMS_OBJECT )
			throw 0;
		return h->val_obj->provideParam(name2);
	}
}


DOMSItemHolder& DOMSObject::provideParam(int number) {
	if( number!=-1 && number < items.size() )
		return *items[number];
	else {
		int n = number;
		int sz = (int) items.size();
		if( n==-1 || n>sz )
			n = sz;
		DOMSItemHolder* h = new DOMSItemHolder(n);
		if( n == sz )
			items.push_back(h);
		else
			items[n] = h;
		return *items[n];
	}
}

DOMSItemHolder* DOMSObject::getParam(const String& name) {
	size_t ppos = name.strpos('.');
	if( ppos==-1 ) {
		if( params.find(name)==params.end() )
			return NULL;
		else
			return params[name];
	} else {
		StringB name1 = name.substr(0,ppos);
		StringB name2 = name.substr(ppos+1);
		DOMSItemHolder* h;
		if( params.find(name1)!=params.end() )
			h = params[name1];
		else
			return 0;
		if( h->type != DOMS_OBJECT )
			throw 0;
		return h->val_obj->getParam(name2);
	}
}

DOMSItemHolder* DOMSObject::getParam(int number) {
	if( number!=-1 && number < items.size() )
		return items[number];
	else
		return NULL;
}

DOMSItem::DOMSItem(DOMSObject& object, const String& name)
		: object(object), name(name), use_name(true) {
}

DOMSItem::DOMSItem(DOMSObject& object, int number)
		: object(object), number(number), use_name(false) {
}

DOMSItem::~DOMSItem() {
}

int DOMSItem::operator= (int val) {
	DOMSItemHolder& h = use_name ? object.provideParam(name) : object.provideParam(number);
	if( !use_name ) number = h.number;
	h.set(val);
	return val;
}

double DOMSItem::operator= (double val) {
	DOMSItemHolder& h = use_name ? object.provideParam(name) : object.provideParam(number);
	if( !use_name ) number = h.number;
	h.set(val);
	return val;
}

//const char* DOMSItem::operator= (const char* val) {
//	*this = String(val);
//	return val;
//}

const String& DOMSItem::operator= (const String& val) {
	DOMSItemHolder& h = use_name ? object.provideParam(name) : object.provideParam(number);
	if( !use_name ) number = h.number;
	h.set(val);
	return val;
}

const DOMSObject& DOMSItem::operator= (const DOMSObject& val) {
	DOMSItemHolder& h = use_name ? object.provideParam(name) : object.provideParam(number);
	if( !use_name ) number = h.number;
	h.set(val);
	return val;
}

void DOMSItem::set_bin(const void* data, size_t size) {
	DOMSItemHolder& h = use_name ? object.provideParam(name) : object.provideParam(number);
	if( !use_name ) number = h.number;
	h.set(data,size);
}

DOMSItem::operator int() const {
	const DOMSItemHolder* h = use_name ? object.getParam(name) : object.getParam(number);
	if( h==0 )
		throw 0;
	if( h->type != DOMS_INT )
		throw 1;
	return h->val_int;
}

DOMSItem::operator DOMSObject& () const {
	const DOMSItemHolder* h = use_name ? object.getParam(name) : object.getParam(number);
	if( h==0 )
		throw 0;
	if( h->type != DOMS_OBJECT )
		throw 1;
	return *h->val_obj;
}

DOMSItem::operator StringB() const {
	const DOMSItemHolder* h = use_name ? object.getParam(name) : object.getParam(number);
	if( h==0 )
		throw 0;
	if( h->type != DOMS_STRING )
		throw 1;
	return *h->val_str;
}

DOMSItem::operator double() const {
	const DOMSItemHolder* h = use_name ? object.getParam(name) : object.getParam(number);
	if( h==0 )
		throw 0;
	if( h->type == DOMS_INT )		// ?
		return h->val_int;
	if( h->type != DOMS_DOUBLE )
		throw 1;
	return h->val_double;
}

void* DOMSItem::get_bin(size_t &size) const {
	const DOMSItemHolder* h = use_name ? object.getParam(name) : object.getParam(number);
	if( h==0 )
		throw 0;
	if( h->type != DOMS_BINARY )
		throw 1;
	size = h->val_size;
	return h->val_data;
}

void DOMSItem::make_object() {
	DOMSItemHolder& h = use_name ? object.provideParam(name) : object.provideParam(number);
	if( !use_name ) number = h.number;
	h.set( DOMSObject() );
}

DOMSObject& DOMSItem::obj() {
	const DOMSItemHolder* h = use_name ? object.getParam(name) : object.getParam(number);
	if( h!=NULL && h->type==DOMS_OBJECT )
		return *h->val_obj;
	make_object();
	return *this;
}

bool DOMSItem::exsists() const {
	const DOMSItemHolder* h = use_name ? object.getParam(name) : object.getParam(number);
	return h!=NULL;
}

const wchar_t* DOMSItem::getname() const {
	if( !use_name )
		throw 0;
	return name.bytes();
}

const int DOMSItem::getnumber() const {
	if( use_name )
		throw 0;
	return number;
}

DOMSType DOMSItem::type() const {
	const DOMSItemHolder* h = use_name ? object.getParam(name) : object.getParam(number);
	return h->type;
}

DOMSItemHolder::DOMSItemHolder(const String& _name) : name(_name), use_name(true) {
	type = DOMS_NULL;
}

DOMSItemHolder::DOMSItemHolder(int number) : number(number), use_name(false) {
	type = DOMS_NULL;
}

DOMSItemHolder::DOMSItemHolder(const DOMSItemHolder& h)
		: name(h.name), number(h.number), use_name(h.use_name) {
	type = h.type;
	switch( type ) {
	case DOMS_INT:
		val_int = h.val_int;
		break;
	case DOMS_DOUBLE:
		val_double = h.val_double;
		break;
	case DOMS_STRING:
		val_str = new StringB(*h.val_str);
		break;
	case DOMS_OBJECT:
		val_obj = new DOMSObject(*h.val_obj);
		break;
	}

}

DOMSItemHolder::~DOMSItemHolder() {
	clear();
}

void DOMSItemHolder::set(int val) {
	if( type != DOMS_INT )
		clear();
	type = DOMS_INT;
	val_int = val;
}

void DOMSItemHolder::set(double val) {
	if( type != DOMS_DOUBLE )
		clear();
	type = DOMS_DOUBLE;
	val_double = val;
}

void DOMSItemHolder::set(const String& val) {
	if( type != DOMS_STRING )
		clear();
	type = DOMS_STRING;
	val_str = new StringB(val);
}

void DOMSItemHolder::set(const DOMSObject& val) {
	if( type != DOMS_OBJECT )
		clear();
	val_obj = new DOMSObject(val);
	type = DOMS_OBJECT;
		// Protection froc recursion, e.g. obj["f"] = obj
		// - field f is not copied (as it is still NULL)
}

void DOMSItemHolder::set(const void* data, size_t size) {
	if( type != DOMS_OBJECT )
		clear();
	type = DOMS_BINARY;
	val_data = new char[size];
	val_size = size;
	memcpy(val_data,data,size);
}

void DOMSItemHolder::clear() {
	switch( type ) {
	case DOMS_STRING:
		delete val_str;
		break;
	case DOMS_OBJECT:
		delete val_obj;
		break;
	case DOMS_BINARY:
		delete val_data;
		break;
	}
	type = DOMS_NULL;
}

uint64_t DOMSItemHolder::size() const {
	uint64_t size = sizeof(FDOMS_ITEM);
	switch( type ) {
	case DOMS_INT:
		size += sizeof(int);
		break;
	case DOMS_DOUBLE:
		size += sizeof(double);
		break;
	case DOMS_STRING:
		size += ALIGN32( 2 * val_str->len() );
		break;
	case DOMS_OBJECT:
		size += val_obj->size();
		break;
	case DOMS_BINARY:
		size += ALIGN32(val_size);
		break;
	}
	if( use_name ) {
		size += ALIGN32( 2 * name.len() );
	}
	return size;
}

uint64_t DOMSItemHolder::store(void* data, uint64_t avail) const {
	uint64_t size = sizeof(FDOMS_ITEM);
	if( size > avail )
		return 0;
	FDOMS_ITEM* o = (FDOMS_ITEM*) data;
	uint64_t ret;
	switch( type ) {
	case DOMS_INT:
		size += sizeof(int);
		if( size > avail )
			return 0;
		*(int*)(o+1) = val_int;
		break;
	case DOMS_DOUBLE:
		size += sizeof(double);
		if( size > avail )
			return 0;
		*(double*)(o+1) = val_double;
		break;
	case DOMS_STRING: {
		size_t len = 2 * val_str->len();
		size += ALIGN32(len);
		if( size > avail )
			return 0;
		o->datasize = len;
		memcpy( o+1, val_str->bytes(), len );
		memset( (char*)(o+1) + len, 0, ALIGN32(len)-len );
		break; }
	case DOMS_OBJECT:
		ret = val_obj->store(o+1,avail-size);
		if( ret==0 )
			return 0;
		o->datasize = ret;
		size += ret;
		break;
	case DOMS_BINARY:
		size += ALIGN32(val_size);
		if( size > avail )
			return 0;
		o->datasize = val_size;
		memcpy( o+1, val_data, val_size );
		break;
	}
	o->subsize = size;
	if( use_name ) {
		char* data1 = (char*)( (size_t)data + size );
		size_t name_len = 2 * name.len();
		size += ALIGN32(name_len);
		if( size > avail )
			return 0;
		memcpy( data1, name.bytes(), name_len );
		memset( data1+name_len, 0, ALIGN32(name_len)-name_len );
		o->szname = (uint32_t) name.len();
		o->number = -1;
	} else {
		o->szname = 0;
		o->number = number;
	}
	o->size   = size;
	o->type   = type;
	o->res0   = 0;
	return size;
}

uint64_t DOMSItemHolder::load(void* data, uint64_t avail) {
	uint64_t size = sizeof(FDOMS_ITEM);
	if( size > avail )
		return 0;
	FDOMS_ITEM* o = (FDOMS_ITEM*) data;
	uint64_t rsize = o->size;
	if( rsize > avail )
		return 0;
	type = (DOMSType) o->type;
	uint64_t ret;
	switch( type ) {
	case DOMS_INT:
		size += sizeof(int);
		if( size > rsize )
			return 0;
		val_int = *(int*)(o+1);
		break;
	case DOMS_DOUBLE:
		size += sizeof(double);
		if( size > rsize )
			return 0;
		val_double = *(double*)(o+1);
		break;
	case DOMS_STRING:
		size = o->subsize;
		if( size > rsize ) {
			type = DOMS_NULL;
			return 0;
		}
		val_str = new StringB( (wchar_t*)(o+1), o->datasize/2 );
		break;
	case DOMS_OBJECT:
		val_obj = new DOMSObject();
		ret = val_obj->load(o+1,o->datasize);
		if( ret==0 ) {
			type = DOMS_NULL;
			delete val_obj;
			return 0;
		}
		size += ret;
		break;
	case DOMS_BINARY:
		val_size = o->datasize;
		size += ALIGN32(val_size);
		if( size > avail ) {
			type = DOMS_NULL;
			return 0;
		}
		val_data = new char[val_size];
		memcpy( val_data, o+1, val_size );
		break;
	}
	size += ALIGN32(2*name.len());
	if( size!=rsize )
		return 0;
	return rsize;
}

DOMSItemHolder* DOMSItemHolder::load2(void* data, uint64_t avail, uint64_t &ret_size) {
	uint64_t size = sizeof(FDOMS_ITEM);
	if( size > avail )
		return 0;
	FDOMS_ITEM* o = (FDOMS_ITEM*) data;
	DOMSItemHolder* h;
	if( o->number==-1 ) {
		wchar_t* namedt = (wchar_t*)( (size_t)data + o->subsize );
		h = new DOMSItemHolder( StringB(namedt,o->szname) );
	} else
		h = new DOMSItemHolder(o->number);
	ret_size = h->load(data,avail);
	if( ret_size==0 ) {
		delete h;
		return 0;
	}
	return h;
}

void DOMSItemHolder::print(FILE* fd, int sublevel) {
	wchar_t* spaces = new wchar_t[sublevel+1];
	for( int i=0; i<sublevel; i++ )
		spaces[i] = ' ';
	//memset(spaces,' ',sublevel);
	spaces[sublevel] = '\0';
	if( use_name )
		fwprintf(fd,L"%s%s = ",spaces,name.bytes());
	else
		fwprintf(fd,L"%s[%d] = ",spaces,number);
	switch( type ) {
	case DOMS_INT:
		fwprintf(fd,L"%d\r\n",val_int);
		break;
	case DOMS_DOUBLE:
		fwprintf(fd,L"%lf\r\n",val_double);
		break;
	case DOMS_STRING:
		fwprintf(fd,L"%s\r\n",val_str->bytes());
		break;
	case DOMS_OBJECT:
		fwprintf(fd,L"\r\n");
		val_obj->print(fd,sublevel);
		break;
	case DOMS_BINARY:
		fwprintf(fd,L"binary (%d bytes)\r\n",val_size);
		break;
	}
}

bool DOMSWrite2(StringBuffer& buffer, String preffix, DOMSObject& obj, DOMSWriteFlags flags) {
	wchar_t buf[64];
	wchar_t buf2[64];
	StringB pref;
	if(preffix==L"")
		pref = L"";
	else
		pref = preffix + L".";
	for(DOMSIterator iter = obj.iter(); iter; iter++ ) {
		DOMSItem item = *iter;
		bool f = false;
		switch( item.type() ) {
		case DOMS_INT:
			swprintf( buf, L"=%d", (int)item );
			f = true;
			break;
		case DOMS_DOUBLE:
			//sprintf( buf, "=%lf", (double)item );
			swprintf( buf, L"=%s", float2strw((double)item,10,4).bytes() );
			f = true;
			break;
		case DOMS_STRING: {
			StringB s = ((StringB)item).escapes(ESCAPE_QUOTE);
			//swprintf( buf, L"=\"%s\"", s.bytes() );
			buffer.append(pref);
			if( item.isnamed() )
				buffer.append(item.getname());
			else {
				swprintf( buf2, L"[%i]", item.getnumber() );
				buffer.append(buf2);
			}
			buffer.append(L"=\"");
			buffer.append(s);
			buffer.append(L"\"\r\n");
			break; }
		case DOMS_OBJECT:
			DOMSWrite2(buffer,pref+item.getname(),item.obj(),flags);
			break;
		case DOMS_BINARY: {
			size_t sz;
			void *data = item.get_bin(sz);
			StringB s = EncodeX16DataW(data,sz);
			buffer.append(pref);
			if( item.isnamed() )
				buffer.append(item.getname());
			else {
				swprintf( buf2, L"[%i]", item.getnumber() );
				buffer.append(buf2);
			}
			buffer.append(L"=#");
			buffer.append(s);
			buffer.append(L"\r\n");
			break; }
		}
		if(f) {
			buffer.append(pref);
			if( item.isnamed() )
				buffer.append(item.getname());
			else {
				swprintf( buf2, L"[%i]", item.getnumber() );
				buffer.append(buf2);
			}
			buffer.append(buf);
			buffer.append(L"\r\n");
		}
	}
	return true;
}

bool csrutils::DOMSWrite(wchar_t*& data, size_t& size, DOMSObject& obj, DOMSWriteFlags flags) {
	StringBuffer buffer;
	if( ! DOMSWrite2(buffer,L"",obj,flags) )
		return false;
	size = buffer.length();
	data = new wchar_t[ size+1 ];
	buffer.copy(data);
	data[size] = '\0';
	return true;
}

bool csrutils::DOMSWriteFree(wchar_t* data, size_t size) {
	delete[] data;
	return true;
}

bool csrutils::DOMSRead(const wchar_t* data, size_t size, DOMSObject& obj, DOMSWriteFlags flags) {
	size_t lstart = 0;
	size_t pos = 0;
	while( pos < size ) {
		for(; pos<size && ( data[pos]=='\r' || data[pos]=='\n' ); pos++ ) { }
		lstart = pos;
		for(; pos<size && data[pos]!='\r' && data[pos]!='\n'; pos++ ) { }
		size_t lend = pos;
		for( pos=lstart; pos<size && data[pos]!='='; pos++ ) { }
		StringB name(data+lstart,pos-lstart);
		pos ++;
		if( data[pos]>='0' && data[pos]<='9' || data[pos]=='-' ) {
			StringB val(data+pos,lend-pos);
			if( val.strpos('.')==-1 && val.strpos('E')==-1 ) {
				int v;
				swscanf(val.bytes(),L"%d",&v);
				obj[name] = v;
			} else {
				double v;
				//sscanf(val.bytes(),"%lf",&v);
				v = str2floatw(val);
				obj[name] = v;
			}
		} else if( data[pos]=='"' ) {
			size_t i;
			for( i=pos+1; i<lend; i++ ) {
				if( data[i]=='\\' ) {
					i ++;
					continue;
				}
				if( data[i]=='"' )
					break;
			}
			StringB val(data+pos+1,i-pos-1);
			StringB v = val.invescapes(ESCAPE_QUOTE);
			obj[name] = v;
		} else if( data[pos]=='#' ) {
			StringB val(data+pos+1,lend-pos-1);
			size_t size = DecodeX16DataW(val,0,0);
			void* data = new char[size];
			size = DecodeX16DataW(val,data,size);
			if( size==0 )
				return false;
			obj[name].set_bin(data,size);
			delete[] data;
		}
		pos = lend;
	}
	return false;
}