/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **            Dynamic array             **
 ** Version 1.0 (2015 DECEMBER 04)       **
 ******************************************/

#pragma once

#include <csrmem.h>

namespace csrutils {

template<typename T> class DynArray {
	T *data;
	size_t cap, size;
public:

	DynArray() {
		data = 0;
		cap = 0;
		size = 0;
	}

	~DynArray() {
		for( size_t i=0; i<size; i++ )
			data[i].~T();
		if(cap)
			heap_free((void*)data);
	}

	void add(const T &o) {
		if( cap > size )
			new ((T*)&data[size++]) T (o);
		else {
			expand();
			add(o);
		}
	}

	void add(T &&o) {
		if( cap > size )
			new ((T*)&data[size++]) T (o);
		else {
			expand();
			add(o);
		}
	}

	void expand() {
		if( cap==0 ) {
			data = heap_alloc<T>();
			cap = 1;
		} else {
			T *newdata = heap_alloc<T>(2*cap);
			for( int i=0; i<size; i++ ) {
				new (&newdata[i]) T(data[i]);
				data[i].~T();
			}
			heap_free((void*)data);
			data = newdata;
			cap = 2*cap;
		}
	}
};

}