/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comsp/csrutils/  **
 ******************************************
 **                 File                 **
 ** Version 1.0 (2015 DECEMBER 04)       **
 ******************************************/

#define _CRT_NON_CONFORMING_SWPRINTFS
#define _CRT_SECURE_NO_WARNINGS

#ifdef _WIN32
#include <Windows.h>
#include <Shlwapi.h>
#endif
#ifdef __unix__
#include <unistd.h>
#endif

#include <string.h>
#include <csrfile.h>
#include "csrfile_int.h"

using namespace csrutils;

/*File::File() {
	ctx = new FileCtx;
	DWORD sz = GetCurrentDirectoryW(0,NULL);
	WCHAR* buf = new WCHAR[sz+1];
	sz = GetCurrentDirectoryW(sz+1,buf);
	ctx->path  = buf;
	ctx->hFile = NULL;
	delete[] buf;
}*/

File::File(FileBase* _base) {
	base = _base->clone();
}

File::File(const File& file) {
	base = file.base->clone();
}

File::File(String path) {
	base = new(heap_alloc<FileBaseDisk>()) FileBaseDisk(path);
}

File::~File() {
	heap_free(base);
}

StringB File::name() const {
	return base->getname();
}

File File::child(String path) const {
	return File( base->getchild(path) );
}

FileIterator File::files() const {
	return FileIterator(*this);
}

//IOStream* File::open() {
//	return base->getstream();
//}

File File::getcurdir() {
#ifdef _WIN32
	DWORD sz = GetCurrentDirectoryW(0,NULL);
	WCHAR* buf = heap_alloc<WCHAR>(sz+1);
	sz = GetCurrentDirectoryW(sz+1,buf);
	File file = File( new(heap_alloc<FileBaseDisk>()) FileBaseDisk(buf) );
	heap_free(buf);
#endif
#ifdef __linux__
	char buf[512];
	getcwd(buf,512);
	StringB bufw = StringA(buf,STRING_UTF8);
	File file = File( new(heap_alloc<FileBaseDisk>()) FileBaseDisk(bufw) );
#endif
	return file;
}

/*File File::getfile(String path) {
	return File( new FileBaseDisk(path) );
}*/







FileIterator::FileIterator(const File& dir) {
	base = dir.getbase()->getfiles();
	curfile = base->getnext();
}

FileIterator::~FileIterator() {
	heap_free(base);
}

FileIterator::operator File() const {
	return File(curfile);
}

void FileIterator::operator ++(int) {
	heap_free(curfile);
	curfile = base->getnext();
}

FileIterator::operator bool() const {
	return curfile != NULL;
}


bool FileBase::copy(const File& dst) {
	char buffer[1024];
	size_t sz;
	FileHandle *s_src = this->open(File_Read);
	if( s_src==NULL )
		return false;
	FileHandle *s_dst = dst.open(File_Write);
	if( s_dst==NULL )
		return false;
	for(;;) {
		s_src->read(buffer,1024,sz);
		if( sz==0 )
			break;
		s_dst->write(buffer,sz,sz);
	}
	heap_free(s_src);
	heap_free(s_dst);
	return true;
}

bool createpath_chain(File& file, Url path) {
	Url parent = path.parent();
	File parent_file = file.getfile(parent);
	if( parent_file.type() == FileType_Directory )
		return true;
	if( !createpath_chain(file,parent) )
		return false;
	if( !parent_file.createdir() )
		return false;
	return true;
}

bool FileBase::createpath() {
	File file(this);
	Url path = getpath();
	return createpath_chain(file,path);
}



inline wchar_t getX16digitw(unsigned char d) {
	if( d<0xA )
		return '0' + d;
	if( d<0x10 )
		return 'A' + (d-0xA);
	return '?';
}

FileMD5::operator StringB() const {
	return EncodeX16DataW(digest,16);
}
