/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comsp/csrutils/  **
 ******************************************
 **                 Url                  **
 ** Version 1.1 (2015 JANUARY 20)        **
 ******************************************/

#define _CRT_NON_CONFORMING_SWPRINTFS
#define _CRT_SECURE_NO_WARNINGS

#include <vector>
#include <list>
#include <wchar.h>

#include <csrurl.h>
#include <csrmacro.h>

using namespace csrutils;

struct UrlNodeA {
	size_t begin;
	size_t end;
};

// ! Warning: str will be changes (separated by nulls to part strings)
inline void url_decompose(wchar_t* str, std::list<String>& snodes) {
	std::vector<UrlNodeA> nodes;
	size_t i, pos = 0;
	for( i=0; str[i]!='\0'; i++ ) {
		if( str[i]=='\\' || str[i]=='/' ) {
			if( i != pos ) {
				UrlNodeA node = { pos, i };
				nodes.push_back(node);
				str[i] = '\0';
			}
			pos = i + 1;
		}
	}
	if( i != pos ) {
		UrlNodeA node = { pos, i };
		nodes.push_back(node);
	}
	for( auto iter=nodes.begin(); iter!=nodes.end(); iter++ )
		snodes.push_back( String( str + (*iter).begin ) );
}

inline StringB url_compose(uint64_t flags, const std::list<String>& snodes) {
	if( flags&URL_WINDOWS ) {
		StringB buf = L"";
		for( auto iter=snodes.begin(); iter!=snodes.end(); iter++ ) {
			if( iter == snodes.begin() )
				buf += *iter;
			else
				buf += String(L"\\") + *iter;
		}
		if( (flags&URL_DIRECTORY) && !(flags&URL_NOENDSLASH) )
			buf += L"\\";
		return buf;
	} else { // if( flags&URL_UNIX ) {
		StringB buf = (flags&URL_ABSOLUTE) ? L"/" : L"";
		for( auto iter=snodes.begin(); iter!=snodes.end(); iter++ ) {
			if( iter == snodes.begin() )
				buf += *iter;
			else
				buf += String(L"/") + *iter;
		}
		if( (flags&URL_DIRECTORY) && !(flags&URL_NOENDSLASH) )
			buf += L"/";
		return buf;
	}
}

void Url::_format(uint64_t _flags, const wchar_t* str0) {
	flags = _flags;
	if( flags & URL_NOROMAT ) {
		string = String(str0);
		flags &= URL_MASK;
		return;
	}
	StringB buf0 = str0;
	wchar_t* str = buf0.bytes();
	if( flags & URL_AUTO ) {
		if( !(flags&URL_WINDOWS) && !(flags&URL_UNIX) ) {
			for( size_t i=0; str[i]!='\0'; i++ ) {
				if( str[i]=='\\' )
					flags |= URL_WINDOWS;
				else if( str[i]=='/' )
					flags |= URL_UNIX;
			}
		}
		if( !(flags&URL_WINDOWS) && !(flags&URL_UNIX) )
			flags |= URL_OS;
	}
	if( !(flags&URL_ABSOLUTE) ) {
		if( (flags&URL_UNIX) && str[0] == '/' )
			flags |= URL_ABSOLUTE;
		else if( (flags&URL_WINDOWS) && str[2] == ':' )
			flags |= URL_ABSOLUTE;
	}
	if( str0[wcslen(str)-1]=='/' || str0[wcslen(str)-1]=='\\' )
		flags |= URL_DIRECTORY;
	std::list<String> snodes;
	url_decompose(str,snodes);
	for( auto iter=snodes.begin(); iter!=snodes.end(); iter++ ) {
		if( *iter == L".." ) {
			if( iter != snodes.begin() ) {
				auto iter0 = iter; iter0--;
				if( iter0 == snodes.begin() ) {
					snodes.erase(iter);
					snodes.erase(iter0);
					iter = snodes.begin();
				} else {
					auto iter1 = iter0; iter1--;
					snodes.erase(iter);
					snodes.erase(iter0);
					iter = iter1;
				}
			}
		}
		if( *iter == L"." ) {
			if( iter == snodes.begin() ) {
				snodes.erase(iter);
				iter = snodes.begin();
			} else {
				auto iter0 = iter; iter0--;
				snodes.erase(iter);
				iter = iter0;
			}
		}
	}
	string = url_compose(flags,snodes);
	flags &= URL_MASK;
}

StringB Url::name() const {
	isize_t i;
	const wchar_t* str = string.bytes();
	size_t end = string.len();
	for( i=end; i>=0; i-- ) {
		if( str[i]=='\\' || str[i]=='/' ) {
			if( i != end-1 )
				break;
			else
				end = i;
		}
	}
	return StringB( str+i+1, end-i );
}

Url Url::child(const Url& url) const {
	uint64_t f = (flags&(~URL_DIRECTORY)) | (url.flags&URL_DIRECTORY);
	return Url( f, string + L"/" + url.string );
}

Url Url::resolve(const Url& url) const {
	if( url.flags & URL_ABSOLUTE )
		return url;
	uint64_t f = (flags&(~URL_DIRECTORY)) | (url.flags&URL_DIRECTORY);
	if( flags & URL_DIRECTORY )
		return Url( f, string + L"/" + url.string );
	else
		return Url( f, (StringB)parent() + url.string );
}

Url Url::parent() const {
	isize_t i;
	const wchar_t* str = string.bytes();
	size_t end = string.len();
	for( i=end; i>=0; i-- ) {
		if( str[i]=='\\' || str[i]=='/' ) {
			if( i != end-1 )
				break;
			else
				end = i;
		}
	}
	i++;
	if( i==0 ) {
		if( flags & URL_ABSOLUTE )
			return Url(flags|URL_DIRECTORY,L"/");
		else
			return Url(flags|URL_DIRECTORY,L".");
	} else {
		Url url;
		url.flags = flags | URL_DIRECTORY;
		url.string = string.substr(0,i-1);
		return url;
	}
}

bool Url::is_parent(const Url& url) const {
	size_t len = url.string.len();
	if( url.string[len-1]=='\\' || url.string[len-1]=='/' )
		len --;
	if( !( string.len() > len+1 ) )
		return false;
	if( string[len]!='\\' && string[len]!='/' )
		return false;
	for( size_t i=0; i<len; i++ ) {
		if( string[i]!=url.string[i] ) {
			if( ( url.string[i]!='\\' && url.string[i]!='/' )
				|| ( string[i]!='\\' && string[i]!='/' ) )
				return false;
		}
	}
	return true;
}

Url Url::relative(const Url& url) const {
	std::list<String> this_nodes;
	std::list<String> url_nodes;
	StringB this_buf = string;
	StringB url_buf = url.string;
	url_decompose(this_buf.bytes(),this_nodes);
	url_decompose(url_buf.bytes(),url_nodes);
	size_t len = MIN( this_nodes.size(), url_nodes.size() );
	auto titer = this_nodes.begin();
	auto uiter = url_nodes.begin();
	while( titer!=this_nodes.end() && uiter!=url_nodes.end() ) {
		if( *titer != *uiter )
			break;
		titer ++;
		uiter ++;
	}
	std::list<String> rel_nodes;
	while( titer!=this_nodes.end() ) {
		rel_nodes.push_back( L".." );
		titer ++;
	}
	while( uiter!=url_nodes.end() ) {
		rel_nodes.push_back( *uiter );
		uiter ++;
	}
	uint64_t f = (flags&URL_MASKPARENT) | (url.flags&URL_MASKCHILD);
	return Url(f|URL_NOROMAT,url_compose(f,rel_nodes));
}
