/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **       String formatting utils        **
 ** Version: 0.1 (2015-08-14)            **
 ******************************************/

#pragma once

#include <csrutils.h>
#include <csrstr.h>
#include <csrstream.h>

namespace csrutils {

class FormatReader;
class FormatWriter;

const uint64_t READER_QUOTED	= 0x0010;

struct FormatReaderCtx;

class FormatReader {
	FormatReaderCtx* ctx;
	uint64_t flags;
	size_t buf_pos;
	size_t buf_end;
	size_t buf_begin;
	wchar_t* buffer;

public:
	FormatReader(const String& buffer);
	FormatReader(IOStream& stream);
	~FormatReader();

	void setflag(uint64_t mask, uint64_t flag);

	String line();
	String word();

	FormatReader& operator>>(int&);
	FormatReader& operator>>(double&);
	FormatReader& operator>>(String&);

	void skip_space();
	void buf_fill();

	bool eof();
};

// TODO: Implement
struct FormatByteWriter_Ctx;
class FormatByteWriter {
	FormatByteWriter_Ctx *ctx;

public:
	FormatByteWriter(IOStream *output);
	~FormatByteWriter();

	int write0(void *data, size_t bytes);

	inline FormatByteWriter& operator<<(int64_t x) {
		write0(&x,sizeof(int64_t));
		return *this;
	}

	FormatByteWriter& operator<<(int x) {
		write0(&x,sizeof(int));
		return *this;
	}

	FormatByteWriter& operator<<(short x) {
		write0(&x,sizeof(int64_t));
		return *this;
	}

	FormatByteWriter& operator<<(char x) {
		write0(&x,sizeof(char));
		return *this;
	}

	FormatByteWriter& operator<<(double x) {
		write0(&x,sizeof(double));
		return *this;
	}

	FormatByteWriter& operator<<(float x) {
		write0(&x,sizeof(float));
		return *this;
	}

	FormatByteWriter& operator<<(const String& s);
};



// TODO: Implement
struct FormatByteReader_Ctx;
class FormatByteReader {
	FormatByteReader_Ctx *ctx;

public:
	FormatByteReader(IOStream *input);
	~FormatByteReader();

	int read0(void *data, size_t bytes);

	inline FormatByteReader& operator>>(int64_t& x) {
		read0(&x,sizeof(int64_t));
		return *this;
	}

	FormatByteReader& operator>>(int& x) {
		read0(&x,sizeof(int));
		return *this;
	}

	FormatByteReader& operator>>(short& x) {
		read0(&x,sizeof(int64_t));
		return *this;
	}

	FormatByteReader& operator<<(char& x) {
		read0(&x,sizeof(char));
		return *this;
	}

	FormatByteReader& operator>>(double& x) {
		read0(&x,sizeof(double));
		return *this;
	}

	FormatByteReader& operator>>(float& x) {
		read0(&x,sizeof(float));
		return *this;
	}

	FormatByteReader& operator>>(String& s);
};

};
