#include <ctype.h>

#include <csrutils.h>
#include <csrformat.h>
#include <csrmacro.h>

using namespace csrutils;

#define BUFFER_SIZE		512

const uint64_t READER_EOF		= 0x0001;
const uint64_t READER_BUFEMPTY	= 0x0002;
const uint64_t READER_STRING	= 0x0004;

struct csrutils::FormatReaderCtx {
	String srcstring;
	IOStream* srcstream;
	size_t srcpos;
};

FormatReader::FormatReader(const String& _buffer) {
	ctx = new FormatReaderCtx;
	flags = READER_STRING | READER_BUFEMPTY;
	ctx->srcstring = _buffer;
	ctx->srcpos    = 0;
	buffer = new wchar_t[BUFFER_SIZE];
	buf_pos   = 0;
	buf_begin = 0;
	buf_end   = 0;
}

FormatReader::~FormatReader() {
	delete[] buffer;
	delete ctx;
}

void FormatReader::setflag(uint64_t mask, uint64_t flag) {
	flags = (flags&(~mask)) | (flag&mask);
}

String FormatReader::line() {
	StringBuffer str;
	return str;
}

FormatReader& FormatReader::operator>>(int& res) {
	skip_space();
	size_t i = 0;
	int b = 0;
	for(;;) {
		for( i=buf_pos; i!=buf_end; i=(i+1)%BUFFER_SIZE ) {
			if( buffer[i]>='0' && buffer[i]<='9' )
				b = b*10 + (buffer[i]-'0');
			else
				goto proc_end;
		}
		buf_fill();
	}
proc_end:
	buf_pos   = i;
	buf_begin = i;
	res = b;
	if( buf_begin==buf_end )
		flags |= READER_BUFEMPTY;
	return *this;
}

FormatReader& FormatReader::operator>>(String& res) {
	StringBuffer str;
	skip_space();
	for(;;) {
		for( ; buf_pos!=buf_end; buf_pos=(buf_pos+1)%BUFFER_SIZE ) {
			if( !isspace(buffer[buf_pos]) )
				str.append(buffer[buf_pos]);
			else
				goto proc_end;
		}
		buf_fill();
	}
proc_end:
	res = str;
	buf_begin = buf_pos;
	if( buf_begin==buf_end )
		flags |= READER_BUFEMPTY;
	return *this;
}

void FormatReader::skip_space() {
	for(;;) {
		for( ; buf_pos!=buf_end; buf_pos=(buf_pos+1)%BUFFER_SIZE ) {
			if( ! isspace(buffer[buf_pos]) )
				return;
		}
		buf_fill();
	}
}

void FormatReader::buf_fill() {
	if( flags & READER_STRING ) {
		size_t len = ctx->srcstring.len();
		size_t aval = len - ctx->srcpos;
		if( aval==0 ) {
			flags |= READER_EOF;
			return;
		}
		if( buf_end == BUFFER_SIZE ) {
			buf_end = 0;
		}
		if( buf_begin < buf_end || (flags&READER_BUFEMPTY) ) {
			size_t sz = MIN(BUFFER_SIZE-buf_end,aval);
			memcpy( buffer+buf_end, ctx->srcstring.bytes()+ctx->srcpos, 2*sz );
			buf_end += sz;
			ctx->srcpos += sz;
			flags &= ~READER_BUFEMPTY;
		} else {
			size_t sz = MIN(buf_begin-buf_end,aval);
			memcpy( buffer+buf_end, ctx->srcstring.bytes()+ctx->srcpos, 2*sz );
			buf_end += sz;
			ctx->srcpos += sz;
			flags &= ~READER_BUFEMPTY;
		}
	}
}
