/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comp/csrutils/   **
 ******************************************
 **            Logging utils             **
 ** Version 0.1 (2015 DECEMBER 04)       **
 ******************************************/

#define _CRT_SECURE_NO_WARNINGS

#include <stdarg.h>
#include <stdio.h>

#include <csrlog.h>
#include <csrstr.h>

using namespace csrutils;

const char *loglvevelnames[] = {
	"INFO ",
	"WARN ",
	"ERROR",
	"FAIL "
};

struct Log_Ctx {
	FILE *fd;
	LogLevel loglevel;
};

Log_Ctx *curlog = 0;

void csrutils::LogInit(const wchar_t *logfile, LogLevel level) {
	Log_Ctx *log = new Log_Ctx;
	curlog = log;
#ifdef _WIN32
	curlog->fd = _wfopen(logfile,L"wt");
#else
	StringAB path = StringAB( String(logfile), STRING_UTF8 );
	curlog->fd = fopen(path.bytes(),"wt");
#endif
	curlog->loglevel = level;
}

void csrutils::LogMsg(uint32_t error_id, LogLevel level, const char *desc, ...) {
	if( curlog==NULL )
		return;
	va_list args;
	va_start(args,desc);
	if( level >= curlog->loglevel ) {
		fprintf(curlog->fd,"%s %4u: ",loglvevelnames[level],error_id);
		vfprintf(curlog->fd,desc,args);
		fprintf(curlog->fd,"\n");
	}
}

void csrutils::LogMsgDebug(uint32_t error_id, LogLevel level, const char *desc, ...) {
#ifndef _DEBUG
	return;
#endif
	if( curlog==NULL )
		return;
	va_list args;
	va_start(args,desc);
	fprintf(curlog->fd,"%s DEBUG %4ud: ",loglvevelnames[level],error_id);
	vfprintf(curlog->fd,desc,args);
	fprintf(curlog->fd,"\n");
}
