#include <csrutils.h>
#include <csrmem.h>
#include <csrmthread.h>

#include <sys/mman.h>


using namespace csrutils;

struct csrutils::MemorySection_Ctx {
	SyncObject sync;
};

MemorySection::MemorySection(size_t size, uint64_t flags) {
	int prot = 0;
	switch( flags & (MEM_READ|MEM_WRITE|MEM_EXEC) ) {
	case 0:
		prot = PROT_NONE;
		break;
	case MEM_READ:
		prot = PROT_READ;
		break;
	case MEM_READ|MEM_WRITE:
		prot = PROT_READ|PROT_WRITE;
		break;
	case MEM_EXEC:
		prot = PROT_EXEC;
		break;
	case MEM_EXEC|MEM_READ:
		prot = PROT_EXEC|PROT_READ;
		break;
	case MEM_EXEC|MEM_READ|MEM_WRITE:
		prot = PROT_EXEC|PROT_READ|PROT_WRITE;
		break;
	}
	ctx = new( heap_alloc<MemorySection_Ctx>() ) MemorySection_Ctx;
	ctx->sync.uses_inc();
	t_size = size;
	t_data = mmap(nullptr,t_size,prot,MAP_ANONYMOUS|MAP_PRIVATE,-1,0);
}

MemorySection::MemorySection(const MemorySection& sec) {
	ctx    = sec.ctx;
	t_data = sec.t_data;
	t_size = sec.t_size;
	ctx->sync.uses_inc();
}

MemorySection::MemorySection(MemorySection&& sec) {
	ctx    = sec.ctx;
	t_data = sec.t_data;
	t_size = sec.t_size;
	sec.ctx    = nullptr;
	sec.t_data = nullptr;
	sec.t_size = 0;
}

MemorySection::~MemorySection() {
	if( ctx != nullptr && ctx->sync.uses_dec() ) {
		heap_free(ctx);
		munmap(t_data,t_size);
	}
	ctx    = nullptr;
	t_data = nullptr;
	t_size = 0;
}

MemorySection& MemorySection::operator= (const MemorySection& sec) {
	if( ctx != nullptr && ctx->sync.uses_dec() ) {
		heap_free(ctx);
		munmap(t_data,t_size);
	}
	ctx    = sec.ctx;
	t_data = sec.t_data;
	t_size = sec.t_size;
	ctx->sync.uses_inc();
	return *this;
}

MemorySection& MemorySection::operator= (MemorySection&& sec) {
	if( ctx != nullptr && ctx->sync.uses_dec() ) {
		heap_free(ctx);
		munmap(t_data,t_size);
	}
	ctx    = sec.ctx;
	t_data = sec.t_data;
	t_size = sec.t_size;
	sec.ctx    = nullptr;
	sec.t_data = nullptr;
	sec.t_size = 0;
	return *this;
}
