#include <csrutils.h>
#include <csrmem.h>
#include <csrmthread.h>

#ifdef _WIN32
#include <windows.h>
#endif

using namespace csrutils;

struct csrutils::MemorySection_Ctx {
	SyncObject sync;
};

MemorySection::MemorySection(size_t size, uint64_t flags, uint64_t prot) {
	DWORD allcflags = MEM_COMMIT;
	DWORD protflags = 0;
	switch( prot ) {
	case 0:
		protflags = PAGE_NOACCESS;
		break;
	case MEM_READ:
		protflags |= PAGE_READONLY;
		break;
	case MEM_READ|MEM_WRITE:
		protflags |= PAGE_READWRITE;
		break;
	case MEM_EXEC:
		protflags |= PAGE_EXECUTE;
		break;
	case MEM_EXEC|MEM_READ:
		protflags |= PAGE_EXECUTE_READ;
		break;
	case MEM_EXEC|MEM_READ|MEM_WRITE:
		protflags |= PAGE_EXECUTE_READWRITE;
		break;
	}
	if( flags & MEM_LARGE )
		allcflags |= MEM_LARGE_PAGES;
	if( flags & MEM_SHARED )
		throw 0;
	ctx = new( heap_alloc<MemorySection_Ctx>() ) MemorySection_Ctx;
	ctx->sync.uses_inc();
	t_size = size;
	t_data = VirtualAlloc(NULL,t_size,allcflags,protflags);
}

MemorySection::MemorySection(const MemorySection& sec) {
	ctx    = sec.ctx;
	t_data = sec.t_data;
	t_size = sec.t_size;
	ctx->sync.uses_inc();
}

MemorySection::MemorySection(MemorySection&& sec) {
	ctx    = sec.ctx;
	t_data = sec.t_data;
	t_size = sec.t_size;
	sec.ctx    = nullptr;
	sec.t_data = nullptr;
	sec.t_size = 0;
}

MemorySection::~MemorySection() {
	if( ctx != nullptr && ctx->sync.uses_dec() ) {
		heap_free(ctx);
		VirtualFree(t_data,t_size,MEM_RELEASE);
	}
	ctx    = nullptr;
	t_data = nullptr;
	t_size = 0;
}

MemorySection& MemorySection::operator= (const MemorySection& sec) {
	if( ctx != nullptr && ctx->sync.uses_dec() ) {
		heap_free(ctx);
		VirtualFree(t_data,t_size,MEM_RELEASE);
	}
	ctx    = sec.ctx;
	t_data = sec.t_data;
	t_size = sec.t_size;
	ctx->sync.uses_inc();
	return *this;
}

MemorySection& MemorySection::operator= (MemorySection&& sec) {
	if( ctx != nullptr && ctx->sync.uses_dec() ) {
		heap_free(ctx);
		VirtualFree(t_data,t_size,MEM_RELEASE);
	}
	ctx    = sec.ctx;
	t_data = sec.t_data;
	t_size = sec.t_size;
	sec.ctx    = nullptr;
	sec.t_data = nullptr;
	sec.t_size = 0;
	return *this;
}

size_t MemorySection::getPageSize() {
	SYSTEM_INFO info;
	GetSystemInfo(&info);
	return info.dwPageSize;
}

size_t MemorySection::getLargePageSize() {
	return GetLargePageMinimum();
}
