#include <Windows.h>

#include "csrmthread.h"

using namespace csrutils;

struct ExecitionMonitor_ThreadCtx {
	HANDLE hThread;
};

struct csrutils::ExecitionMonitor_Ctx {
	volatile bool sign_abort;
	volatile int proc_count;
	HANDLE ev_aborted;
	HANDLE ev_running;
};

ExecitionMonitor::ExecitionMonitor() {
	ctx = new ExecitionMonitor_Ctx;
	ctx->sign_abort = false;
	ctx->proc_count = 0;
	ctx->ev_aborted = CreateEvent(0,true,false,0);
	ctx->ev_running = CreateEvent(0,true,true,0);
}

ExecitionMonitor::~ExecitionMonitor() {
	CloseHandle(ctx->ev_aborted);
	delete ctx;
}


void ExecitionMonitor::abort() {
	if( ctx->proc_count > 0 ) {
		ctx->sign_abort = true;
		WaitForSingleObject(ctx->ev_aborted,-1);
		ctx->sign_abort = false;
	}
}

void ExecitionMonitor::pause() {
	ResetEvent(ctx->ev_running);
}

void ExecitionMonitor::unpause() {
	SetEvent(ctx->ev_running);
}

void ExecitionMonitor::proc_attend(bool& aborted) {
	if( ctx->sign_abort )
		aborted=true;
	else
		aborted=false;
	WaitForSingleObject(ctx->ev_running,-1);
}

void ExecitionMonitor::proc_enter() {
	ctx->proc_count ++;
}

void ExecitionMonitor::proc_exit() {
	ctx->proc_count --;
	if( ctx->proc_count==0 && ctx->sign_abort ) {
		SetEvent(ctx->ev_aborted);
	}
}
