/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 **            Inner header file         **
 ** Version: 0.1 (2015-08-09)            **
 ******************************************/

#pragma once

#ifdef __linux__
#include <pthread.h>
#endif

#ifdef _WIN32
#include <windows.h>
#endif

namespace csrutils {

class ThreadBase {
	SyncObject syncobj;
#ifdef _WIN32
	HANDLE hThread;
	HANDLE hStart;
	HANDLE hComplete;
#endif
#ifdef __linux__
	pthread_t pThread;
	SyncEvent* pStart;
	SyncEvent* pComplete;
#endif
	uint32_t threadId;
	ThreadJob* runjob;

public:
	ThreadBase(ThreadJob* runjob);
	~ThreadBase();

	void start();
	void proc();
	void join();

	ThreadBase(const Thread& t) = delete;
	void operator=(const ThreadBase&) = delete;

	static void sleep(int time);

	friend class Thread;
};

#ifdef _WIN32

#define interlocked_increment(x)		InterlockedIncrement(x)
#define interlocked_decrement(x)		InterlockedDecrement(x)

#define thread_event_t					HANDLE
#define thread_event_init(event,mr,is)	(*event) = CreateEventW(NULL,mr,is,NULL)
#define thread_event_set(event)			SetEvent((*event))
#define thread_event_reset(event)		ResetEvent((*event))
#define thread_event_wait(event)		WaitForSingleObject((*event),INFINITE)

#define thread_mutex_t					CRITICAL_SECTION
#define thread_mutex_init(mutex)		InitializeCriticalSection(mutex)
#define thread_mutex_deinit(mutex)		DeleteCriticalSection(mutex)
#define thread_mutex_lock(mutex)		EnterCriticalSection(mutex)
#define thread_mutex_unlock(mutex)		LeaveCriticalSection(mutex)

#define thread_cond_t					CONDITION_VARIABLE
#define thread_cond_init(cond)			InitializeConditionVariable(cond)
#define thread_cond_deinit(cond)
#define thread_cond_wait(cond,mutex)	SleepConditionVariableCS(cond,mutex,INFINITE)
#define thread_cond_signal(cond)		WakeAllConditionVariable(cond)

#endif

#ifdef __linux__

#define interlocked_increment(x)		__sync_add_and_fetch(x,1)
#define interlocked_decrement(x)		__sync_sub_and_fetch(x,1)

#define thread_event_t					PSyncEvent
#define thread_event_init(event,mr,is)	*(event) = new SyncEvent(mr,is)
#define thread_event_set(event)			(*(event))->set()
#define thread_event_reset(event)		(*(event))->reset()
#define thread_event_wait(event)		(*(event))->wait()

#define thread_mutex_t					pthread_mutex_t
#define thread_mutex_init(mutex)	
#define thread_mutex_deinit(mutex)	
#define thread_mutex_lock(mutex)		pthread_mutex_lock(mutex)
#define thread_mutex_unlock(mutex)		pthread_mutex_unlock(mutex)

#define thread_cond_t					pthread_cond_t
#define thread_cond_init(cond)			
#define thread_cond_deinit(cond)
#define thread_cond_wait(cond,mutex)	pthread_cond_wait(cond,mutex)
#define thread_cond_signal(cond)		pthread_cond_signal(cond)

#endif

};
