/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 ** SyncEvent linux implementation       **
 ** Version: 0.1 (2015-08-09)            **
 ******************************************/

#include <pthread.h>
#include <stdio.h>

#include <csrmthread.h>

#define INFINITE -1

using namespace csrutils;

struct csrutils::CSRSYNCEVENT {
	pthread_cond_t cond;
	pthread_mutex_t mutex;
	volatile bool state;
	bool auto_res;
};

SyncEvent::SyncEvent(bool manualReset, bool initState) {
	nevent = new CSRSYNCEVENT;
	nevent->state = initState;
	nevent->auto_res = ! manualReset;
	//printf("Create new SyncEvent(%i,%i)\n",(int)manualReset,(int)initState);
}

SyncEvent::~SyncEvent() {
	delete nevent;
}

void SyncEvent::set() {
	pthread_mutex_lock(&nevent->mutex);
	nevent->state = true;
	pthread_cond_signal(&nevent->cond);
	pthread_mutex_unlock(&nevent->mutex);
}

void SyncEvent::reset() {
	nevent->state = false;
}

void SyncEvent::wait() {
	//WaitForSingleObject(nevent,INFINITE);
	//printf("Waiting for SyncEvent(%i,%i)\n",(int)nevent->auto_res,(int)nevent->state);
	if( !nevent->state ) {
		pthread_mutex_lock(&nevent->mutex);
		while( ! nevent->state )
			pthread_cond_wait(&nevent->cond,&nevent->mutex);
		if( nevent->auto_res )
			nevent->state = false;
		pthread_mutex_unlock(&nevent->mutex);
	}
	//printf("Waited for SyncEvent(%i,%i)\n",(int)nevent->auto_res,(int)nevent->state);
}

void SyncEvent::wait(uint32_t time) {
	//WaitForSingleObject(nevent,time);
	// TODO:
	throw "TODO: void SyncEvent::wait(uint32_t time)";
}

void SyncEvent::waitAll(SyncEvent* events, size_t count) {
	waitAll(events,count,INFINITE);
}

void SyncEvent::waitAll(SyncEvent* events, size_t count, uint32_t time) {
	/*HANDLE *hevents = new HANDLE[count];
	for( size_t i=0; i<count; i++ )
		hevents[i] = events[i].nevent;
	WaitForMultipleObjects((DWORD)count,hevents,TRUE,time);
	delete[] hevents;*/
}

void SyncEvent::waitSingle(SyncEvent* events, size_t count) {
	waitSingle(events,count,INFINITE);
}

void SyncEvent::waitSingle(SyncEvent* events, size_t count, uint32_t time) {
	/*HANDLE *hevents = new HANDLE[count];
	for( size_t i=0; i<count; i++ )
		hevents[i] = events[i].nevent;
	WaitForMultipleObjects((DWORD)count,hevents,FALSE,time);
	delete[] hevents;*/
}
