/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 ** SyncEvent windows implementation     **
 ** Version: 0.1 (2015-08-09)            **
 ******************************************/

#include <windows.h>

#include <csrmthread.h>

using namespace csrutils;

SyncEvent::SyncEvent(bool manualReset, bool initState) {
	nevent = CreateEvent(NULL,manualReset,initState,NULL);
}

SyncEvent::~SyncEvent() {
	CloseHandle(nevent);
}

void SyncEvent::set() {
	SetEvent(nevent);
}

void SyncEvent::reset() {
	ResetEvent(nevent);
}

void SyncEvent::wait() {
	WaitForSingleObject(nevent,INFINITE);
}

void SyncEvent::wait(uint32_t time) {
	WaitForSingleObject(nevent,time);
}

void SyncEvent::waitAll(SyncEvent* events, size_t count) {
	waitAll(events,count,INFINITE);
}

void SyncEvent::waitAll(SyncEvent* events, size_t count, uint32_t time) {
	HANDLE *hevents = new HANDLE[count];
	for( size_t i=0; i<count; i++ )
		hevents[i] = events[i].nevent;
	WaitForMultipleObjects((DWORD)count,hevents,TRUE,time);
	delete[] hevents;
}

void SyncEvent::waitSingle(SyncEvent* events, size_t count) {
	waitSingle(events,count,INFINITE);
}

void SyncEvent::waitSingle(SyncEvent* events, size_t count, uint32_t time) {
	HANDLE *hevents = new HANDLE[count];
	for( size_t i=0; i<count; i++ )
		hevents[i] = events[i].nevent;
	WaitForMultipleObjects((DWORD)count,hevents,FALSE,time);
	delete[] hevents;
}
