/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 ** SyncObject windows implementation    **
 ** Version: 0.1 (2015-08-09)            **
 ******************************************/

#include <windows.h>

#include <csrutils.h>
#include <csrmthread.h>
#include <csrmem.h>

using namespace csrutils;

struct csrutils::SyncObject_Ctx {
	volatile size_t counter;
	CRITICAL_SECTION mutex;
};

SyncObject::SyncObject() {
	ctx = heap_alloc<SyncObject_Ctx>();
	ctx->counter = 0;
	InitializeCriticalSection(&ctx->mutex);
}

SyncObject::~SyncObject() {
	DeleteCriticalSection(&ctx->mutex);
	heap_free(ctx);
}

void SyncObject::uses_inc() {
	InterlockedIncrement(&ctx->counter);
}

bool SyncObject::uses_dec() {
	InterlockedDecrement(&ctx->counter);
	return ctx->counter == 0;
}

size_t SyncObject::uses() {
	return ctx->counter;
}

void SyncObject::lock() {
	EnterCriticalSection(&ctx->mutex);
}

void SyncObject::unlock() {
	LeaveCriticalSection(&ctx->mutex);
}
