/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 ** SyncQueue linux implementation       **
 ** Version: 0.1 (2015-08-09)            **
 ******************************************/

#include <pthread.h>
#include <stdio.h>

#include <csrmthread.h>

using namespace csrutils;

struct csrutils::SyncQueueBase_Ctx {
	bool b_full;
	bool b_empty;
	pthread_cond_t s_empty;
	pthread_cond_t s_full;
	pthread_mutex_t s_mutex;
};

SyncQueueBase::SyncQueueBase(size_t _maxcount) {
	maxcount = _maxcount;
	ctx = new SyncQueueBase_Ctx;
	p_front = 0;
	p_back  = 0;
	p_count = 0;
	p_data  = new void*[maxcount];
}

SyncQueueBase::~SyncQueueBase() {
	// TODO: provide no waiters (i.e. all threads have finished)
	
	delete ctx;
}

void SyncQueueBase::push(void* obj) {
	//WaitForSingleObject(ctx->s_full,INFINITE);
	//EnterCriticalSection(&ctx->s_mutex);
	pthread_mutex_lock(&ctx->s_mutex);
	while( ctx->b_full )
		pthread_cond_wait(&ctx->s_full,&ctx->s_mutex);
	p_data[p_back++] = obj;
	p_back = p_back % maxcount;
	if( p_count==0 ) {
		ctx->b_empty = false;
		pthread_cond_signal(&ctx->s_empty);
	}
	p_count ++;
	if( p_count==maxcount )
		ctx->b_full = true;
	//LeaveCriticalSection(&ctx->s_mutex);
	//ReleaseSemaphore(ctx->s_empty,1,NULL);
	pthread_mutex_unlock(&ctx->s_mutex);
}

void* SyncQueueBase::pop() {
	//WaitForSingleObject(ctx->s_empty,INFINITE);
	//EnterCriticalSection(&ctx->s_mutex);
	pthread_mutex_lock(&ctx->s_mutex);
	while( ctx->b_empty )
		pthread_cond_wait(&ctx->s_empty,&ctx->s_mutex);
	void* obj = p_data[p_front++];
	p_front = p_front % maxcount;
	if( p_count==maxcount ) {
		ctx->b_full = false;
		pthread_cond_signal(&ctx->s_full);
	}
	p_count --;
	if( p_count==0 )
		ctx->b_empty = true;
	//LeaveCriticalSection(&ctx->s_mutex);
	//ReleaseSemaphore(ctx->s_full,1,NULL);
	pthread_mutex_unlock(&ctx->s_mutex);
	return obj;
}
