/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 ** SyncQueue windows implementation     **
 ** Version: 0.1 (2015-08-09)            **
 ******************************************/

#include <windows.h>

#include <csrmthread.h>

using namespace csrutils;

struct csrutils::SyncQueueBase_Ctx {
	HANDLE s_empty;
	HANDLE s_full;
	CRITICAL_SECTION s_mutex;
};

SyncQueueBase::SyncQueueBase(size_t _maxcount) {
	maxcount = _maxcount;
	ctx = new SyncQueueBase_Ctx;
	ctx->s_empty = CreateSemaphore(NULL,0,(int)maxcount,NULL);
	ctx->s_full  = CreateSemaphore(NULL,(int)maxcount,(int)maxcount,NULL);
	InitializeCriticalSection(&ctx->s_mutex);
	p_front = 0;
	p_back  = 0;
	p_count = 0;
	p_data  = new void*[maxcount];
}

SyncQueueBase::~SyncQueueBase() {
	// TODO: provide no waiters (i.e. all threads have finished)
	DeleteCriticalSection(&ctx->s_mutex);
	CloseHandle(ctx->s_full);
	CloseHandle(ctx->s_empty);
	delete ctx;
}

void SyncQueueBase::push(void* obj) {
	WaitForSingleObject(ctx->s_full,INFINITE);
	EnterCriticalSection(&ctx->s_mutex);
	p_data[p_back++] = obj;
	p_back = p_back % maxcount;
	LeaveCriticalSection(&ctx->s_mutex);
	ReleaseSemaphore(ctx->s_empty,1,NULL);
}

void* SyncQueueBase::pop() {
	WaitForSingleObject(ctx->s_empty,INFINITE);
	EnterCriticalSection(&ctx->s_mutex);
	void* obj = p_data[p_front++];
	p_front = p_front % maxcount;
	LeaveCriticalSection(&ctx->s_mutex);
	ReleaseSemaphore(ctx->s_full,1,NULL);
	return obj;
}
