/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 ** Thread linux implementation          **
 ** Version: 0.1 (2015-08-09)            **
 ******************************************/

#include <pthread.h>
#include <unistd.h>
#include <time.h>
#include <vector>

#include <csrmthread.h>
#include "csrmthread_int.h"

using namespace csrutils;

void* CsrThreadProc(void* p);

class ThreadItem {
public:
	pthread_t pThread;
	ThreadBase* pThreadBase;

	inline ThreadItem(pthread_t t, ThreadBase *b) : pThread(t), pThreadBase(b) { }
	inline bool operator==(ThreadItem t2) const {
		return 0 != pthread_equal(pThread,t2.pThread);
	}
};

std::vector<ThreadItem> csr_thread_map1;

ThreadBase::ThreadBase(ThreadJob* _runjob) {
	pthread_attr_t attr;
	runjob    = _runjob;
	pStart    = new SyncEvent(true,false);
	pComplete = new SyncEvent(true,false);
	pthread_attr_init(&attr);
	if( pthread_create(&pThread,&attr,CsrThreadProc,this)!=0 )
		throw 0;
	pthread_attr_destroy(&attr);
	threadId  = 0;
	csr_thread_map1.push_back( ThreadItem(pThread,this) );
}

ThreadBase::~ThreadBase() {
	delete pStart;
	delete pComplete;
}

void ThreadBase::start() {
	if( runjob!=NULL )
		pStart->set();
}

void ThreadBase::proc() {
	pStart->wait();
	runjob->run();
	pComplete->set();
}

void ThreadBase::join() {
	pComplete->wait();
	//void *ret;
	//pthread_join(pThread,&ret);
}

void ThreadBase::sleep(int time) {
	struct timespec req;
	struct timespec rem;
	req.tv_sec  = time / 1000;
	req.tv_nsec = ( time % 1000 ) * 1000000;
	//::sleep(time);
	nanosleep(&req,&rem);
}

void* CsrThreadProc(void* p) {
	ThreadBase* thread = (ThreadBase*) p;
	Thread th(thread);
	thread->proc();
	//delete thread;
	return 0;
}

Thread::Thread() {
	base = NULL;
}

Thread::Thread(ThreadJob* runjob) {
	base = new ThreadBase(runjob);
	base->syncobj.uses_inc();
}


Thread::Thread(ThreadBase* _base) {
	base = _base;
	if( base!=NULL )
		base->syncobj.uses_inc();
}

Thread::Thread(const Thread& t) {
	base = t.base;
	if( base!=NULL )
		base->syncobj.uses_inc();
}

Thread::~Thread() {
	if( base!=NULL && base->syncobj.uses_dec() ) {
		delete base;
	}
}

Thread& Thread::operator= (const Thread& t) {
	if( base!=NULL && base->syncobj.uses_dec() ) {
		delete base;
	}
	base = t.base;
	if( base!=NULL )
		base->syncobj.uses_inc();
	return *this;
}

bool Thread::operator==(const Thread& t) const {
	return base == t.base;
}

void Thread::start() {
	base->start();
}

void Thread::join() {
	base->join();
}

void Thread::sleep(int time) {
	ThreadBase::sleep(time);
}

Thread Thread::current() {
	pthread_t thread = pthread_self();
	ThreadItem it(thread,NULL);
	for( auto iter=csr_thread_map1.begin(); iter!=csr_thread_map1.end(); iter++ ) {
		if( it==(*iter) )
			return Thread( (*iter).pThreadBase );
	}
	return Thread();
}

