/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 ** Thread windows implementation        **
 ** Version: 0.1 (2015-08-09)            **
 ******************************************/

#include <windows.h>
#include <map>

#include <csrmthread.h>
#include "csrmthread_int.h"

using namespace csrutils;

DWORD WINAPI CsrThreadProc(LPVOID p);

std::map<uint32_t,ThreadBase*> csr_thread_map1;

ThreadBase::ThreadBase(ThreadJob* _runjob) {
	runjob    = _runjob;
	hStart    = CreateEvent(NULL,true,false,NULL);
	hComplete = CreateEvent(NULL,true,false,NULL);
	hThread   = CreateThread(0,0,CsrThreadProc,this,0,0);
	threadId  = GetThreadId(hThread);
	csr_thread_map1[threadId] = this;
}

ThreadBase::~ThreadBase() {
	csr_thread_map1.erase(threadId);
	CloseHandle(hThread);
	CloseHandle(hStart);
	CloseHandle(hComplete);
}

void ThreadBase::start() {
	if( runjob!=NULL )
		SetEvent(hStart);
}

void ThreadBase::join() {
	WaitForSingleObject(hThread,INFINITE);
	//WaitForSingleObject(hComplete,INFINITE);
}

void ThreadBase::proc() {
	WaitForSingleObject(hStart,INFINITE);
	runjob->run();
	SetEvent(hComplete);
}

void ThreadBase::sleep(int time) {
	Sleep(time);
}

DWORD WINAPI CsrThreadProc(LPVOID p) {
	ThreadBase* thread = (ThreadBase*) p;
	Thread th(thread);
	thread->proc();
	//delete thread;
	return 0;
}

Thread::Thread() {
	base = NULL;
}

Thread::Thread(ThreadJob* runjob) {
	base = new ThreadBase(runjob);
	base->syncobj.uses_inc();
}


Thread::Thread(ThreadBase* _base) {
	base = _base;
	if( base!=NULL )
		base->syncobj.uses_inc();
}

Thread::Thread(const Thread& t) {
	base = t.base;
	if( base!=NULL )
		base->syncobj.uses_inc();
}

Thread::~Thread() {
	if( base!=NULL && base->syncobj.uses_dec() ) {
		delete base;
	}
}

Thread& Thread::operator= (const Thread& t) {
	if( base!=NULL && base->syncobj.uses_dec() ) {
		delete base;
	}
	base = t.base;
	if( base!=NULL )
		base->syncobj.uses_inc();
	return *this;
}

bool Thread::operator==(const Thread& t) const {
	return base == t.base;
}

void Thread::start() {
	base->start();
}

void Thread::join() {
	base->join();
}

void Thread::sleep(int time) {
	Sleep(time);
}

Thread Thread::current() {
	uint32_t threadId = GetCurrentThreadId();
	if( csr_thread_map1.find(threadId) != csr_thread_map1.end() )
		return Thread( csr_thread_map1[threadId] );
	else
		return Thread();
}

