/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **          String functions            **
 ** Version 1.1 (2016 MARCH 27)          **
 ******************************************/

#pragma once

#include <csrutils.h>
#include <csrmem.h>

namespace csrutils {

/*
Table of String classes
Bits	Char. represent.	w/buffer	w/o buffer
16-bit	Unicode				String		StringB
8-bit	UTF8				StringU		StringUB
8-bit	CodePage			StringA		StringAB
*/

class String;
class StringB;
class StringA;
class StringAB;
class StringU;
class StringUB;

enum StringEscapeFlags {
	ESCAPE_QUOTE = 0x0001
};

CSRUTILS_DEFINE_ENUM_FLAG_OPERATORS(StringEscapeFlags);

enum StringEncoding {
	STRING_UTF8,
	STRING_CP1251
};

inline StringEncoding get_default_encoding() {
	return STRING_CP1251;
}

}

#include <csrstringw.h>
#include <csrstringu.h>
#include <csrstringa.h>

namespace csrutils {

class StringBuffer {
	wchar_t *buffer;
	size_t capacity;
	size_t size;

public:
	StringBuffer();
	~StringBuffer();

	operator StringB();
	
	bool append(wchar_t ch);
	bool append(String str);
	bool append(const wchar_t *str);
	size_t length();

	// returns pointer to chars
	// might be invalidated by further calls
	wchar_t *getdata();

	// copies chars into buffer (must be allocated)
	void copy(wchar_t *buffer);
};

StringAB EncodeX16DataA(const void* data, size_t length);
StringB EncodeX16DataW(const void* data, size_t length);

// if data=0 or length=0 doesn't perform conversion,
// returns requires length of data to be allocated (i.e. str.len/2)
// arg 'length' must be set as size of allocated buffer
size_t DecodeX16DataA(const StringA& str, void* data, size_t length);
size_t DecodeX16DataW(const String& str, void* data, size_t length);

StringA float2str(double num, int digits, int minexp, bool endzeros=false);
double str2float(const StringA& str);
StringB int2strw(int64_t num, int digits=0);
StringB float2strw(double num, int digits, int minexp, bool endzeros=false);
double str2floatw(const String& str);

}