/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comp/csrutils/   **
 ******************************************
 **          String functions            **
 ** Version 1.1 (2016 MARCH 27)          **
 ******************************************/

#define _CRT_SECURE_NO_WARNINGS

#include <string.h>
#include <wchar.h>

#include <csrutils.h>
#include <csrstr.h>
#include <csrosutils.h>

using namespace csrutils;

StringB String::escapes(StringEscapeFlags flags) {
	size_t i;
	size_t size = tlen;
	for( i=0; i<tlen; i++ ) {
		wchar_t c = tdata[i];
		if( c=='"' && flags&ESCAPE_QUOTE || c=='\\' )
			size ++;
	}
	StringB res;
	res.tdata = heap_alloc<wchar_t>(size+1);
	res.tlen  = size;
	size = 0;
	wchar_t* d = res.tdata;
	for( i=0; i<tlen; i++ ) {
		wchar_t c = tdata[i];
		if( c=='"' && flags&ESCAPE_QUOTE ) {
			d[size++] = '\\';
			d[size++] = '\"';
		} else if( c=='\\' ) {
			d[size++] = '\\';
			d[size++] = '\\';
		} else
			d[size++] = c;
	}
	d[size] = '\0';
	return res;
}

StringB String::invescapes(StringEscapeFlags flags) {
	size_t i;
	size_t size = tlen;
	for( i=0; i<tlen; i++ ) {
		wchar_t c = tdata[i];
		if( c=='\\' ) {
			i ++;
			size --;
		}
	}
	StringB res;
	res.tdata = heap_alloc<wchar_t>(size+1);
	res.tlen  = size;
	size = 0;
	wchar_t* d = res.tdata;
	for( i=0; i<tlen; i++ ) {
		wchar_t c = tdata[i];
		if( c=='\\' ) {
			d[size++] = tdata[i+1];
			i ++;
		} else
			d[size++] = c;
	}
	d[size] = '\0';
	return res;
}

StringB::StringB(const StringA& str) {
	tdata = os_convertencoding2( str.bytes(), str.getencoding() );
	tlen  = wcslen(tdata);
}

String& StringB::operator= (const StringA& str) {
	if( tdata!=nullptr )
		heap_free(tdata);
	tdata = os_convertencoding2( str.bytes(), str.getencoding() );
	tlen  = wcslen(tdata);
	return *this;
}

inline wchar_t getX16digitw(unsigned char d) {
	if( d<0xA )
		return '0' + d;
	if( d<0x10 )
		return 'A' + (d-0xA);
	return '?';
}

inline unsigned char getX16digitvalw(wchar_t d) {
	if( d>='0' && d<= '9' )
		return d - '0';
	if( d>='A' && d<='F' )
		return 0xA + (d-'A');
	if( d>='a' && d<='f' )
		return 0xA + (d-'a');
	return -1;
}

StringB csrutils::EncodeX16DataW(const void* data, size_t length) {
	const unsigned char* cdata = (const unsigned char*) data;
	wchar_t *str = new wchar_t[2*length];
	for( int i=0; i<length; i++ ) {
		str[2*i+0] = getX16digitw(cdata[i]>>4);
		str[2*i+1] = getX16digitw(cdata[i]&0xF);
	}
	StringB s(str,2*length);
	delete[] str;
	return s;
}

size_t csrutils::DecodeX16DataW(const String& str, void* data, size_t length) {
	if( str.len() % 2 != 0 )
		return 0;
	if( data==0 || length==0 )
		return str.len() / 2;
	if( length < str.len() / 2 )
		return 0;
	size_t len = str.len() / 2;
	unsigned char* cdata = (unsigned char*) data;
	const wchar_t *sstr = str.bytes();
	for( int i=0; i<len; i++ ) {
		unsigned char d1 = getX16digitvalw(sstr[2*i+0]);
		unsigned char d2 = getX16digitvalw(sstr[2*i+1]);
		if( d1==-1 || d2==-1 )
			return 0;
		cdata[i] = (d1<<4) | d2;
	}
	return len;
}
