/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comp/csrutils/   **
 ******************************************
 **          String functions            **
 ** Version 1.1 (2016 MARCH 27)          **
 ******************************************/

#define _CRT_SECURE_NO_WARNINGS

#include <string.h>
#include <stdio.h>

#include <csrutils.h>
#include <csrstr.h>
#include <csrosutils.h>

using namespace csrutils;

StringAB StringA::escapes(StringEscapeFlags flags) {
	size_t i;
	size_t size = tlen;
	for( i=0; i<tlen; i++ ) {
		char c = tdata[i];
		if( c=='"' && flags&ESCAPE_QUOTE || c=='\\' )
			size ++;
	}
	StringAB res;
	res.tdata = heap_alloc<char>(size+1);
	res.tlen  = size;
	size = 0;
	char* d = res.tdata;
	for( i=0; i<tlen; i++ ) {
		char c = tdata[i];
		if( c=='"' && flags&ESCAPE_QUOTE ) {
			d[size++] = '\\';
			d[size++] = '\"';
		} else if( c=='\\' ) {
			d[size++] = '\\';
			d[size++] = '\\';
		} else
			d[size++] = c;
	}
	d[size] = '\0';
	return res;
}

StringAB StringA::invescapes(StringEscapeFlags flags) {
	size_t i;
	size_t size = tlen;
	for( i=0; i<tlen; i++ ) {
		char c = tdata[i];
		if( c=='\\' ) {
			i ++;
			size --;
		}
	}
	StringAB res;
	res.tdata = heap_alloc<char>(size+1);
	res.tlen  = size;
	size = 0;
	char* d = res.tdata;
	for( i=0; i<tlen; i++ ) {
		char c = tdata[i];
		if( c=='\\' ) {
			d[size++] = tdata[i+1];
			i ++;
		} else
			d[size++] = c;
	}
	d[size] = '\0';
	return res;
}

StringAB::StringAB(const String& str, StringEncoding enc) {
	tdata = os_convertencoding1( str.bytes(), enc );
	tlen  = strlen(tdata);
}

StringA& StringAB::operator= (const String& str) {
	if( tdata!=nullptr )
		heap_free(tdata);
	tdata = os_convertencoding1( str.bytes(), tenc );
	tlen  = strlen(tdata);
	return *this;
}

inline char getX16digit(unsigned char d) {
	if( d<0xA )
		return '0' + d;
	if( d<0x10 )
		return 'A' + (d-0xA);
	return '?';
}

inline char getX16digitval(unsigned char d) {
	if( d>='0' && d<= '9' )
		return d - '0';
	if( d>='A' && d<='F' )
		return 0xA + (d-'A');
	if( d>='a' && d<='f' )
		return 0xA + (d-'a');
	return -1;
}

StringAB csrutils::EncodeX16DataA(const void* data, size_t length) {
	const unsigned char* cdata = (const unsigned char*) data;
	char *str = new char[2*length];
	for( int i=0; i<length; i++ ) {
		str[2*i+0] = getX16digit(cdata[i]>>4);
		str[2*i+1] = getX16digit(cdata[i]&0xF);
	}
	StringAB s = StringAB(str,2*length);
	delete[] str;
	return s;
}

size_t csrutils::DecodeX16DataA(const StringA& str, void* data, size_t length) {
	if( str.len() % 2 != 0 )
		return 0;
	if( data==0 || length==0 )
		return str.len() / 2;
	if( length < str.len() / 2 )
		return 0;
	size_t len = str.len() / 2;
	unsigned char* cdata = (unsigned char*) data;
	const char *sstr = str.bytes();
	for( int i=0; i<len; i++ ) {
		unsigned char d1 = getX16digitval(sstr[2*i+0]);
		unsigned char d2 = getX16digitval(sstr[2*i+1]);
		if( d1==-1 || d2==-1 )
			return 0;
		cdata[i] = (d1<<4) | d2;
	}
	return len;
}
