/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comp/csrutils/   **
 ******************************************
 **          String functions            **
 ** Version 1.0 (2015 DECEMBER 04)       **
 ******************************************/

#define _CRT_SECURE_NO_WARNINGS
#define _USE_MATH_DEFINES

#include <cmath>

#include "csrutils.h"
#include "csrstr.h"
#include "csrmacro.h"

using namespace csrutils;

int int2charsw(wchar_t* buf, int64_t num, int digits, bool setplus) {
	if( num==0 ) {
		buf[0] = '0';
		return 1;
	}
	char b[32];
	size_t i = 0;
	size_t j = 0;
	if( num<0 ) {
		num = -num;
		buf[j++] = '-';
	} else if(setplus)
		buf[j++] = '+';
	while( num > 0 ) {
		b[ i++ ] = (char)( '0' + (num%10) );
		num /= 10;
	}
	for( size_t k=1; k<=i; k++ )
		buf[j++] = b[i-k];
	return (int) j;
}

int int2digitsw(wchar_t* buf, int64_t num) {
	if( num < 0 )
		num = -num;
	size_t i = 0;
	while( num > 0 ) {
		buf[ i++ ] = (char)( '0' + (num%10) );
		num /= 10;
	}
	return (int) i;
}

StringB csrutils::int2strw(int64_t num, int digits) {
	wchar_t buf[32];
	int j = int2charsw(buf,num,digits,false);
	buf[j] = '\0';
	return StringB(buf);
}

StringB csrutils::float2strw(double num, int digits, int minexp, bool endzeros) {
	if( digits<=0 || digits>50 || minexp<0 || minexp>110 ) {
		throw 0;
	}
	if( isnan(num) )
		return L"NaN";
	if( isinf(num) )
		return L"INF";
	isize_t nsign = SIGN(num);
	double nabs = ABS(num);
	isize_t ndec = 0;
	double man = nabs;
	while( man > 10.0 ) {
		man /= 10.0;
		ndec ++;
	}
	while( man < 1.0 && man > 0.0 ) {
		man *= 10.0;
		ndec --;
	}
	for( int k=1; k<digits; k++ )
		man *= 10.0;
	int64_t iman = (int64_t) man;
	wchar_t buf[128];
	wchar_t buf2[128];
	if( iman==0 ) {
		if( !endzeros || digits==1 )
			return L"0";
		buf[0] = '0';
		buf[1] = '.';
		for( size_t i=1; i<(size_t)digits; i++ )
			buf[i+1] = '0';
		buf[digits+1] = '\0';
		return buf;
	}
	isize_t bufpos = 0;
	isize_t digs = int2digitsw(buf2,iman);
	if( digs > digits ) {
		isize_t shift = digs - digits;
		for( isize_t k=0; k<digits; k++ )
			buf2[k] = buf2[k+shift];
		digs = digits;
		ndec += shift;
	}
	while( digs < digits )
		buf[ digs++ ] = '0';
	if( nsign < 0 )
		buf[ bufpos++ ] = '-';
	isize_t decint = 1;
	if( ABS(ndec) < minexp ) {
		decint += ndec;
		ndec = 0;
	}
	if( decint <= 0 )
		buf[ bufpos++ ] = '0';
	while( decint > 0 && digs>0 ) {
		buf[ bufpos++ ] = buf2[ --digs ];
		decint --;
	}
	while( decint > 0 ) {
		buf[ bufpos++ ] = '0';
		decint --;
	}
	isize_t mindigs = 0;
	if( !endzeros ) {
		while( buf2[mindigs]=='0' && mindigs<digs )
			mindigs ++;
	}
	if( digs > mindigs )
		buf[ bufpos++ ] = '.';
	while( decint < 0 ) {
		buf[ bufpos++ ] = '0';
		decint ++;
	}
	while( digs > mindigs )
		buf[ bufpos++ ] = buf2[ --digs ];
	if( ndec!=0 || minexp==0 ) {
		buf[bufpos++] = 'E';
		bufpos += int2charsw( buf+bufpos, ndec, 0, true );
	}
	buf[bufpos++] = '\0';
	return buf;
}

double csrutils::str2floatw(const String& str) {
	if( str == L"NaN" )
		return NAN;
	if( str == L"INF" )
		return INFINITY;
	isize_t sign = 1;
	double mant = 0.0;
	double mantfrc = 0.1;
	isize_t expn = 0;
	isize_t sign_expn = 1;
	const wchar_t* sstr = str.bytes();
	size_t pos = 0;
	if( sstr[pos]=='-' ) {
		sign = -1;
		pos ++;
	}
	while( sstr[pos]!='\0' && sstr[pos]!='.' && sstr[pos]!='E' ) {
		if( sstr[pos]<'0' || sstr[pos]>'9' )
			return NAN;
		mant = 10.0 * mant + ( sstr[ pos++ ] - '0' );
	}
	if( sstr[pos] == '\0' )
		return mant*sign;
	if( sstr[pos] == '.' ) {
		pos++;
		while( sstr[pos]!='\0' && sstr[pos]!='E' ) {
			if( sstr[pos]<'0' || sstr[pos]>'9' )
				return NAN;
			mant = mant + mantfrc * ( sstr[ pos++ ] - '0' );
			mantfrc /= 10.0;
		}
	}
	if( sstr[pos] == '\0' )
		return mant*sign;
	if( sstr[pos] == 'E' ) {
		pos++;
		if( sstr[pos]=='-' ) {
			sign_expn = -1;
			pos ++;
		} else if( sstr[pos]=='+' )
			pos ++;
		while( sstr[pos]!='\0' ) {
			if( sstr[pos]<'0' || sstr[pos]>'9' )
				return NAN;
			expn = 10 * expn + ( sstr[ pos++ ] - '0' );
		}
		expn *= sign_expn;
		while( expn > 0 ) {
			mant *= 10.0;
			expn --;
		}
		while( expn < 0 ) {
			mant /= 10.0;
			expn ++;
		}
	}
	return mant*sign;
}
