#include <wchar.h>

#include <csrtime.h>

using namespace csrutils;

Time::Time(int year, int month, int day) {
	TimeInfo info = { year, month, day, 0, 0, 0, 0 };
	*this = Time(info);
}

Time::Time(int year, int month, int day, int H, int M, int S) {
	TimeInfo info = { year, month, day, H, M, S, 0 };
	*this = Time(info);
}

Time::Time(TimeInfo info) {
	int Y, Y_1;
	if( info.year > 0 ) {
		Y   = info.year + 1000000;
		Y_1 = info.year +  999999;
	} else {
		Y   = info.year + 1000001;
		Y_1 = info.year + 1000000;
	}
	uint64_t day = ((365*4+1)*100-3) * ( Y_1 / 400 );
	int Y2 = Y_1 % 400;
	day += (365*100+24) * ( Y2 / 100 );
	Y2 = Y2 % 100;
	day += (365*4+1) * (Y2/4) + 365*(Y2%4);
	if( info.month > 2 && (Y%4)==0 && ( (Y%100)!=0 || (Y%400)==0) )
		day += 1;
	if( info.month==2 )
		day += 31;
	else if( info.month==3 )
		day += 59;
	else if( info.month==4 )
		day += 90;
	else if( info.month==5 )
		day += 120;
	else if( info.month==6 )
		day += 151;
	else if( info.month==7 )
		day += 181;
	else if( info.month==8 )
		day += 212;
	else if( info.month==9 )
		day += 243;
	else if( info.month==10 )
		day += 273;
	else if( info.month==11 )
		day += 304;
	else if( info.month==12 )
		day += 334;
	day += info.day - 1;
	base = 86400000L * day + 3600000 * info.hour + 60000 * info.minute
		+ 1000 * info.second + info.millisecond;
}

Time::operator TimeInfo() const {
	TimeInfo info;
	uint64_t time = base % 86400000L;
	int day = (int)( base / 86400000L );
	info.year = (int)( 400 * ( day / ((365*4+1)*100-3) ) );
	day = day % ((365*4+1)*100-3);
	if( day!=146096 ) {
		info.year += (int)( 100 * ( day / ( 365*100 + 24 ) ) );
		day = day % ( 365*100 + 24 );
		info.year += (int)( 4 * ( day / (365*4+1) ) );
		day = day % (365*4+1);
		int dy;
		info.year += dy = (int)( day/365 > 3 ? 3 : day/365 );
		day -= 365 * dy;
	} else {
		info.year += 399;
		day -= 145731;
	}
	if( info.year >= 1000000 )
		info.year -= 999999;
	else
		info.year -= 1000000;
	//info.year ++;
	bool year_ext_feb = false;
	if( (info.year%4)==0 && ( (info.year%100)!=0 || (info.year%400)==0) ) {
		year_ext_feb = day == 59;
		if( day > 59 )
			day -= 1;
	}
	if( day < 31 ) {
		info.month = 1;
		info.day = day + 1;
	} else if( day < 59 || year_ext_feb ) {
		info.month = 2;
		info.day = day - 31 + 1;
	} else if( day < 90 ) {
		info.month = 3;
		info.day = day - 59 + 1;
	} else if( day < 120 ) {
		info.month = 4;
		info.day = day - 90 + 1;
	} else if( day < 151 ) {
		info.month = 5;
		info.day = day - 120 + 1;
	} else if( day < 181 ) {
		info.month = 6;
		info.day = day - 151 + 1;
	} else if( day < 212 ) {
		info.month = 7;
		info.day = day - 181 + 1;
	} else if( day < 243 ) {
		info.month = 8;
		info.day = day - 212 + 1;
	} else if( day < 273 ) {
		info.month = 9;
		info.day = day - 243 + 1;
	} else if( day < 304 ) {
		info.month = 10;
		info.day = day - 273 + 1;
	} else if( day < 334 ) {
		info.month = 11;
		info.day = day - 304 + 1;
	} else {
		info.month = 12;
		info.day = day - 334 + 1;
	} 
	info.hour = (int)( time / 3600000 );
	time -= 3600000 * info.hour;
	info.minute = (int)( time / 60000 );
	time -= 60000 * info.minute;
	info.second = (int)( time / 1000 );
	time -= 1000 * info.second;
	info.millisecond = (int) time;
	return info;
}

Time::operator String() const {
	TimeInfo info = *this;
	wchar_t buffer[256];
#ifdef _WIN32
	swprintf_s(buffer,L"%d.%02d.%02d %02d:%02d:%02d.%03d", info.year, info.month,
		info.day, info.hour, info.minute, info.second, info.millisecond );
#else
	swprintf(buffer,256,L"%d.%02d.%02d %02d:%02d:%02d.%03d", info.year, info.month,
		info.day, info.hour, info.minute, info.second, info.millisecond );
#endif
	return String(buffer);
}

TimeInfo Time::toLocal() const {
	return toLocal( TimeZone::getLocal() );
}

TimeInfo Time::toLocal(const TimeZone& tz) const {
	return TimeInfo( *this + tz.offsetUTC(*this) );
}

Time Time::fromLocal(TimeInfo& info) {
	return fromLocal( info, TimeZone::getLocal() );
}

Time Time::fromLocal(TimeInfo& info, const TimeZone& tz) {
	Time lc = info;
	return lc - tz.offsetLocal(lc);
}

String Time::toLocalString() const {
	return toLocalString( TimeZone::getLocal() );
}

String Time::toLocalString(const TimeZone& tz) const {
	return String( *this + tz.offsetUTC(*this) );
}

Time Time::fromWindows(uint64_t b) {
	b = b / 10000 + 31607443123200000;
	return Time(b);
}



Time TimeDiff::add(Time t) const {
	if( use_info ) {
		TimeDiff base = TimeInfo(t);
		base += *this;
		return base;
	} else {
		return Time( t.base + base );
	}
}

TimeDiff TimeDiff::operator+(const TimeDiff& d) const {
	if( use_info && d.use_info ) {
		TimeInfo inf = d.info;
		inf.year	+= info.year;
		inf.month	+= info.month;
		inf.day		+= info.day;
		inf.hour	+= info.hour;
		inf.minute	+= info.minute;
		inf.second	+= info.second;
		inf.millisecond += info.millisecond;
		return inf;
	} else if( !use_info && !d.use_info ) {
		return this->toComplex() + d;
	} else if( !use_info && d.use_info ) {
		return this->toComplex() + d;
	} else { //if( use_info && !d.use_info ) {
		return *this + d.toComplex();
	}
}

TimeDiff TimeDiff::operator-() const {
	if( use_info ) {
		TimeInfo inf;
		inf.year	= -info.year;
		inf.month	= -info.month;
		inf.day		= -info.day;
		inf.hour	= -info.hour;
		inf.minute	= -info.minute;
		inf.second	= -info.second;
		inf.millisecond	= - info.millisecond;
		return inf;
	} else
		return TimeDiff( -base );
}

TimeDiff TimeDiff::toComplex() const {
	if( use_info )
		return *this;
	int64_t time = base;
	TimeInfo info;
	info.millisecond = (int)( time % 1000 );
	time /= 1000;
	info.second = (int)( time % 60 );
	time /= 60;
	info.minute = (int)( time % 60 );
	time /= 60;
	info.hour = (int)( time % 60 );
	time /= 60;
	info.day = (int)( time % 24 );
	time /= 24;
	info.year = (int)( time / 365 );
	info.month = (int)( time / 30 );
	return info;
}

TimeDiff TimeDiff::toSimple() const {
	int64_t time = ( ( ( (int64_t) ( info.year*365 + info.month*30 + info.day ) * 24L
		+ info.hour ) * 60L + info.minute ) * 60L + info.second ) * 1000L + info.millisecond;
	return TimeDiff(time);
}
