#ifdef _WIN32
#include <Windows.h>
#endif

#include <csrtime.h>

using namespace csrutils;

struct csrutils::TimeZoneCtx {
	virtual TimeZoneCtx* clone() const = 0;
	virtual TimeDiff offsetUTC(Time) const = 0;
	virtual TimeDiff offsetLocal(Time) const = 0;
};

class TimeZoneSimple : public TimeZoneCtx {
	TimeDiff offset;

public:
	TimeZoneSimple(int hours, int minutes) {
		offset = hours * 3600000L + minutes * 60000L;
	}

	TimeZoneCtx* clone() const {
		return new TimeZoneSimple(*this);
	}
	
	TimeDiff offsetUTC(Time) const {
		return offset;
	}

	TimeDiff offsetLocal(Time) const {
		return offset;
	}
};

TimeZone::TimeZone(int hours, int minutes) {
	ctx = new TimeZoneSimple(hours,minutes);
};

TimeZone::TimeZone(const TimeZone& tz) {
	ctx = tz.ctx->clone();
};

TimeZone::~TimeZone() {
	delete ctx;
}

TimeDiff TimeZone::offsetUTC(Time t) const {
	return ctx->offsetUTC(t);
}

TimeDiff TimeZone::offsetLocal(Time t) const {
	return ctx->offsetLocal(t);
}

TimeZone TimeZone::getLocal() {
#ifdef _WIN32
	TIME_ZONE_INFORMATION tzinfo;
	GetTimeZoneInformation(&tzinfo);
	return TimeZone(0,-tzinfo.Bias);
#else
	return TimeZone(0,0);	// TODO:
#endif
}
