/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **              URL utils               **
 ** Version 1.1 (2015 JANUARY 20)        **
 ******************************************/

#pragma once

#define _CSRUTILS_FILE_

#include <csrutils.h>
#include <csrstr.h>

// Formatting and url-classes
#define URL_SCHEME		0x0001L		// http, ftp, file, etc. "https://www.dcsr.ru/"
#define URL_UNIX		0x0002L		// "/home/csr"
#define URL_WINDOWS		0x0004L		// "C:\WINDOWS\System32"
#ifdef _WIN32
	#define URL_OS		0x0004L
#else
	#define URL_OS		0x0002L
#endif

// Formatting and properties
#define URL_ABSOLUTE	0x0100L
#define URL_DIRECTORY	0x0200L
#define URL_NOENDSLASH	0x0400L

#define URL_AUTO		0x10000L
#define URL_NOROMAT		0x20000L

#define URL_MASK		0xFFFFL
#define URL_MASKCHILD	URL_DIRECTORY
#define URL_MASKPARENT	(URL_MASK^URL_DIRECTORY)
#define URL_MASKOS		0x000FL

namespace csrutils {

class Url {
	uint64_t flags;
	StringB string;
	void _format(uint64_t flags, const wchar_t* str);

public:

	inline Url() : flags(0) { }
	inline Url(uint64_t _flags) : flags(_flags) { }
	inline Url(const String& str) : flags(0), string(str) {
		_format(URL_AUTO,str.bytes());
	}
	inline Url(uint64_t _flags, const String& str) : flags(_flags) {
		_format(_flags,str.bytes());
	}
	inline Url(const wchar_t* str) : flags(0), string(str) {
		_format(URL_AUTO,str);
	}
	inline Url(uint64_t _flags, const wchar_t* str) : flags(_flags) {
		_format(_flags,str);
	}

	inline operator StringB () const {
		return string;
	}

	inline void format(uint64_t flags) {
		_format(flags,string.bytes());
	}

	inline void format(uint64_t _flags, uint64_t mask) {
		_format( (flags&(~mask)) | (_flags&mask), string.bytes() );
	}

	StringB name() const;
	inline StringB path() const {
		return string;
	}

	Url child(const Url& url) const;
	// Expand relative `url` using `this` as a base
	Url resolve(const Url& url) const;
	// Make relative url for `url` using `this` as a base
	Url relative(const Url& url) const;

	Url parent() const;
	Url commont_parent(const Url& url) const;

	// Check if `url` is parent of `this`
	bool is_parent(const Url& url) const;
	// Check if `url` is child of `this`
	inline bool is_child(const Url& url) const {
		return url.is_parent(*this);
	}
};

}
