/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comp/csrutils/   **
 ******************************************
 **            Basic header              **
 **   Version: 0.2 (2015-05-25)          **
 ******************************************/

#pragma once

#define _CSRUTILS_

#include <string.h>
#include <stdint.h>
#include <inttypes.h>
//#include <array>

#ifdef _WIN32

#ifndef _CSRLIB_
#define CSRLIB extern "C" _declspec(dllimport)
#define CSRLIBP _declspec(dllimport)
#else
#define CSRLIB extern "C" _declspec(dllexport)
#define CSRLIBP _declspec(dllexport)
#endif

#endif

#ifdef __linux__

#ifndef _CSRLIB_
#define CSRLIB extern "C" __attribute__ ((visibility ("default"))) //__attribute__ ((dllimport))
#define CSRLIBP __attribute__ ((visibility ("default"))) //__attribute__ ((dllimport))
#else
#define CSRLIB extern "C" __attribute__ ((visibility ("default"))) //__attribute__ ((dllexport))
#define CSRLIBP __attribute__ ((visibility ("default"))) //__attribute__ ((dllexport))
#endif

#endif

//typedef int64_t __int64;

namespace csrutils {

class SyncObject;
class HandledObjectSync;

// Object should be allocated with 'new'
class HandledObjectSync {
	SyncObject *sync;

public:
	HandledObjectSync();
	virtual ~HandledObjectSync();

	void addHandle();
	void delHandle();
};

};

#define CSRUTILS_DEFINE_ENUM_FLAG_OPERATORS(ENUMTYPE) \
inline ENUMTYPE operator | (ENUMTYPE a, ENUMTYPE b) { return ENUMTYPE(((int)a) | ((int)b)); } \
inline ENUMTYPE &operator |= (ENUMTYPE &a, ENUMTYPE b) { return (ENUMTYPE &)(((int &)a) |= ((int)b)); } \
inline ENUMTYPE operator & (ENUMTYPE a, ENUMTYPE b) { return ENUMTYPE(((int)a) & ((int)b)); } \
inline ENUMTYPE &operator &= (ENUMTYPE &a, ENUMTYPE b) { return (ENUMTYPE &)(((int &)a) &= ((int)b)); } \
inline ENUMTYPE operator ~ (ENUMTYPE a) { return ENUMTYPE(~((int)a)); } \
inline ENUMTYPE operator ^ (ENUMTYPE a, ENUMTYPE b) { return ENUMTYPE(((int)a) ^ ((int)b)); } \
inline ENUMTYPE &operator ^= (ENUMTYPE &a, ENUMTYPE b) { return (ENUMTYPE &)(((int &)a) ^= ((int)b)); } 
//inline operator ENUMTYPE ( int a ) { return (ENUMTYPE) a; }

template<typename T> class SimpleList {
	T* list;
	size_t count;
public:
	inline SimpleList(size_t _count) {
		count = _count;
		list = new T[count];
	}
	inline ~SimpleList() {
		delete[] list;
	}
	inline const T& operator[](size_t i) const { return list[i]; }
	inline T& operator[](size_t i) { return list[i]; }
	inline operator T*() const { return list; }
	inline size_t size() { return count; }
};

//#ifdef __unix__
//typedef int64_t __int64;
//#endif

#if defined(_WIN64)|defined(__X86_64__)
typedef __int64 isize_t;
#else
typedef int isize_t;
#endif

template<typename T> T* cmalloc() {
	return new T;
}

template<typename T> T* cmalloc(size_t count) {
	return new T[count];
}

template<typename T> void cmfree(T* t) {
	delete t;
}

template<typename T> void cmfreec(T* t) {
	delete[] t;
}

template<typename T> void copy_array(const T* src_begin, const T* src_end, T* dst_begin) {
	while( src_begin!=src_end ) {
		*dst_begin = *src_begin;
		src_begin ++;
		dst_begin ++;
	}
}
