#include <Windows.h>

#include "csrmthread.h"

using namespace csrutils;

struct JobQueue_Item {
	JobQueue_Item *next, *prev;
	JobQueue_Task task;
};

struct csrutils::JobQueue_Ctx {
	int threads_cnt;
	HANDLE *threads;
	JobQueue_Item *first, *last;
	HANDLE hevent, hmutex;
};

DWORD WINAPI JobQueue_ThreadProc(LPVOID lp);

JobQueue::JobQueue(int threads) {
	ctx = new JobQueue_Ctx;
	ctx->threads     = new HANDLE[threads];
	ctx->threads_cnt = threads;
	ctx->first = 0;
	ctx->last  = 0;
	ctx->hmutex = CreateMutex(0,false,0);
	ctx->hevent = CreateEvent(0,false,false,0);
	for( int i=0; i<threads; i++ )
		ctx->threads[i] = CreateThread(0,0,JobQueue_ThreadProc,this,0,0);
}

void JobQueue::push(JobQueue_Task task) {
	WaitForSingleObject(ctx->hmutex,-1);
	JobQueue_Item *item = new JobQueue_Item;
	item->task = task;
	item->next = 0;
	item->prev = ctx->last;
	ctx->last  = item;
	if( item->prev )
		item->prev->next = item;
	else
		ctx->first = item;
	SetEvent(ctx->hevent);
	ReleaseMutex(ctx->hmutex);
}

bool JobQueue::pop(JobQueue_Task &task) {
	for(;;) {
		WaitForSingleObject(ctx->hmutex,-1);
		if( ctx->first!=0 ) {
			JobQueue_Item *item = ctx->first;
			task = item->task;
			ctx->first = item->next;
			if( item->next )
				item->next->prev = 0;
			else
				ctx->last = 0;
			ReleaseMutex(ctx->hmutex);
			delete item;
			return true;
		}
		ReleaseMutex(ctx->hmutex);
		WaitForSingleObject(ctx->hevent,-1);
	}
}

void JobQueue::clear() {
	WaitForSingleObject(ctx->hmutex,-1);
	JobQueue_Item *item = ctx->first;
	while( item ) {
		Objmon_free(item->task.data);
		JobQueue_Item *tmp = item->next;
		delete item;
		item = tmp;
	}
	ctx->first = 0;
	ctx->last  = 0;
	ReleaseMutex(ctx->hmutex);
}

DWORD WINAPI JobQueue_ThreadProc(LPVOID lp) {
	JobQueue *jque = (JobQueue*) lp;
	for(;;) {
		JobQueue_Task task;
		if( !jque->pop(task) )
			return 0;
		task.proc(&task);
		Objmon_free(task.data);
	}
}