#include <Windows.h>

#include "csrutils.h"
#include "csrosutils.h"
#include "csrmem.h"
#include "csrstr.h"

using namespace csrutils;

struct OS_FILE {
	HANDLE hFile;
};

bool os_fopen(OS_FILE **file, const char *path, int flags) {
	*file = new OS_FILE;
	(*file)->hFile = CreateFileA(path,GENERIC_READ,0,0,0,0,0);
	return true;
}

void* os_virtualalloc(size_t size) {
	return VirtualAlloc(NULL,size,MEM_COMMIT,PAGE_READWRITE);
}

void os_virtualfree(void* ref, size_t size) {
	VirtualFree(ref,size,MEM_FREE);
}

inline int get_os_encoding(StringEncoding encoding) {
	switch(encoding) {
	case STRING_UTF8:
		return CP_UTF8;
	case STRING_CP1251:
		return 1251;
	default:
		return 1251;
	}
}

char *os_convertencoding1(const wchar_t* str, StringEncoding encoding) {
	UINT enc = get_os_encoding(encoding);
	int sz = WideCharToMultiByte(enc,0,str,-1,0,0,0,0);
	char* buf = new char[sz+1];
	sz = WideCharToMultiByte(enc,0,str,-1,buf,sz+1,0,0);
	return buf;
}

wchar_t *os_convertencoding2(const char* str, StringEncoding encoding) {
	UINT enc = get_os_encoding(encoding);
	int sz = MultiByteToWideChar(enc,0,str,-1,NULL,0);
	wchar_t* buf = csrutils::heap_alloc<wchar_t>(sz+1);
	sz = MultiByteToWideChar(enc,0,str,-1,buf,sz+1);
	return buf;
}
