#define _CRT_SECURE_NO_WARNINGS
//#define _CSR_INTERNAL_

//#include <Windows.h>
//#include <iostream>
#include <string>
#include <vector>

#include "csrzip.h"
#include "csrzip_int.h"


ZipArchive::ZipArchive(File file, ZipArchiveMode mode) {
	ctx = new ZipArchive_Ctx;
	ctx->file = new File(file);
	ctx->zip_mode = mode;
	switch(mode) {
	case ZIP_READ:
		ctx->file_handle = file.open(File_Read);
		break;
	case ZIP_EDIT:
		ctx->file_handle = file.open(File_Read|File_Write);
		break;
	case ZIP_NEW:
		ctx->file_handle = file.open(File_Write);
		ctx->end_of_data = 0;
		break;
	default:
		throw 0;
	}
	if( mode==ZIP_READ || mode==ZIP_EDIT ) {
		size_t sz;
		FileHandle* fh = ctx->file_handle;
		//fh->setposition( fh->getsize() - sizeof(ZIP_CENTDIREND) );
		fh->pos() = fh->getsize() - sizeof(ZIP_CENTDIREND);
		fh->read( &ctx->hdr_cntdir, sizeof(ZIP_CENTDIREND), sz );
		if( ctx->hdr_cntdir.sign != ZIP_CENTDIREND_SIGN )
			throw 0;
		//fh->setposition( ctx->hdr_cntdir.cdoffset );
		fh->pos() = ctx->hdr_cntdir.cdoffset;
		ctx->end_of_data = ctx->hdr_cntdir.cdoffset;
		size_t readed_sz = 0;
		while( readed_sz < ctx->hdr_cntdir.cdsize ) {
			ZIP_CENTDIRFILE hdr_cntdirfile;
			fh->read( &hdr_cntdirfile, sizeof(ZIP_CENTDIRFILE), sz );
			readed_sz += sz;
			if( hdr_cntdirfile.sign != ZIP_CENTDIRFILE_SIGN )
				break;
			ZIP_FILE* file = new ZIP_FILE;
			file->hdr_cntdirfile = hdr_cntdirfile;
			char* file_path = new char[hdr_cntdirfile.lenname+1];
			fh->read( file_path, hdr_cntdirfile.lenname, sz );
			readed_sz += sz;
			file_path[hdr_cntdirfile.lenname] = '\0';
			file->path = StringA(file_path).convert(STRING_UTF8);
			file->url  = file->path;
			// Skip extra and comm to next header
			//fh->setposition( fh->getposition() + hdr_cntdirfile.lenextr + hdr_cntdirfile.lencomm );
			//fh->pos() += hdr_cntdirfile.lenextr + hdr_cntdirfile.lencomm;
			file->extra = new char[hdr_cntdirfile.lenextr];
			fh->read( file->extra, hdr_cntdirfile.lenextr, sz );
			readed_sz += sz;
			char* file_comment = new char[hdr_cntdirfile.lencomm+1];
			fh->read( file_comment, hdr_cntdirfile.lencomm, sz );
			readed_sz += sz;
			file_comment[hdr_cntdirfile.lencomm] = '\0';
			file->comment = StringA(file_comment).convert(STRING_UTF8);
			ctx->files.push_back(file);
		}
	}
}

ZipArchive::~ZipArchive() {
	if( ctx->zip_mode==ZIP_EDIT || ctx->zip_mode==ZIP_NEW ) {
		/*size_t endoff = 0;
		for( ZIP_FILE* file: ctx->files ) {
			size_t offset = file->hdr_cntdirfile.offset + sizeof(ZIP_LOCALFILEHDR)
				+ file->hdr_cntdirfile.lenname + file->hdr_cntdirfile.lenextr;
			if( offset > endoff )
				endoff = offset;
		}*/
		size_t cdoff = ctx->end_of_data;
		size_t count = 0;
		size_t sz;
		ctx->file_handle->setposition(cdoff);
		for( ZIP_FILE* file: ctx->files ) {
			ctx->file_handle->write(&file->hdr_cntdirfile,sizeof(ZIP_CENTDIRFILE),sz);
			ctx->file_handle->write(file->path.utf8().bytes(),file->hdr_cntdirfile.lenname,sz);
			ctx->file_handle->write(file->extra,file->hdr_cntdirfile.lenextr,sz);
			ctx->file_handle->write(file->comment.utf8().bytes(),file->hdr_cntdirfile.lencomm,sz);
			count ++;
		}
		ZIP_CENTDIREND *hdrcdend = new ZIP_CENTDIREND;
		hdrcdend->sign			= ZIP_CENTDIREND_SIGN;
		hdrcdend->disknum		= 0;
		hdrcdend->diskcdstart	= 0;
		hdrcdend->countcddisk	= (DWORD) count;
		hdrcdend->countcd		= (DWORD) count;
		hdrcdend->cdsize		= (DWORD)( ctx->file_handle->getposition() - cdoff );
		hdrcdend->cdoffset		= (DWORD) cdoff;
		hdrcdend->lencomm		= 0;
		ctx->file_handle->write(hdrcdend,sizeof(ZIP_CENTDIREND),sz);
		ctx->file_handle->setend();
		delete hdrcdend;
	}
	delete ctx->file_handle;
	delete ctx;
}

File ZipArchive::root() {
	return new FileBaseZip(this,L"");
}

File ZipArchive::file(String path) {
	return new FileBaseZip(this,path);
}

/*
bool ZipArchive::open(const char *file, ZipArchiveMode mode) {
	ctx->zip_mode	= mode;
	bool readheader = false;
	switch(mode) {
	case ZIP_READ:		// Open file for reading only
		ctx->zip_file = CreateFileA(file,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,0,0);
		if( ctx->zip_file == INVALID_HANDLE_VALUE )
			return false;
		readheader = true;
		break;
	case ZIP_EDIT:		// Open file for editing (reading and writing)
		ctx->zip_file = CreateFileA(file,GENERIC_READ|GENERIC_WRITE,0,0,OPEN_ALWAYS,0,0);
		if( ctx->zip_file == INVALID_HANDLE_VALUE )
			return false;
		readheader = true;
		break;
	case ZIP_NEW:		// Create and open file for writing
		ctx->zip_file = CreateFileA(file,GENERIC_WRITE,0,0,CREATE_ALWAYS,0,0);
		if( ctx->zip_file == INVALID_HANDLE_VALUE )
			return false;
		break;
	}
		// Read zip header ZIP_CENTDIREND
	if(readheader) {
		DWORD sz;
			// Read ZIP_CENTDIREND header
		SetFilePointer(ctx->zip_file,-(long)sizeof(ZIP_CENTDIREND),0,FILE_END);
		ReadFile(ctx->zip_file,&ctx->hdr_cntdir,sizeof(ZIP_CENTDIREND),&sz,0);
			// Read ZIP_CENTDIRFILE headers
		SetFilePointer(ctx->zip_file,ctx->hdr_cntdir.cdoffset,0,FILE_BEGIN);
		DWORD readed_sz = 0;
		while( readed_sz < ctx->hdr_cntdir.cdsize ) {
			ZIP_CENTDIRFILE hdr_cntdirfile;
			ReadFile(ctx->zip_file,&hdr_cntdirfile,sizeof(ZIP_CENTDIRFILE),&sz,0);
			readed_sz += sz;
			if( hdr_cntdirfile.sign != ZIP_CENTDIRFILE_SIGN )
				break;
			ZipArchiveEntry_Ctx *entry_ctx = new ZipArchiveEntry_Ctx;
			entry_ctx->hdr_cntdirfile = hdr_cntdirfile;
			char *file_path = new char[hdr_cntdirfile.lenname];
			ReadFile(ctx->zip_file,file_path,hdr_cntdirfile.lenname,&sz,0);
			readed_sz += sz;
			entry_ctx->path = std::string(file_path,hdr_cntdirfile.lenname);
			delete[] file_path;
			// Skip extra and comm to next header
			SetFilePointer(ctx->zip_file,hdr_cntdirfile.lenextr+hdr_cntdirfile.lencomm,0,FILE_CURRENT);
			readed_sz += hdr_cntdirfile.lenextr+hdr_cntdirfile.lencomm;
			ctx->files.push_back( new ZipArchiveEntry(this,entry_ctx) );
		}
	}
	return true;
}

bool ZipArchive::close() {
	if( ctx->zip_mode == ZIP_NEW ) {
		DWORD sz;
		DWORD off1 = SetFilePointer(ctx->zip_file,0,0,FILE_CURRENT);

		ZIP_FILE *file = ctx->file_first;
		WORD count = 0;
		while(file) {
			WriteFile(ctx->zip_file,file->d,sizeof(ZIP_CENTDIRFILE),&sz,0);
			WriteFile(ctx->zip_file,file->name,(uint32_t)strlen(file->name),&sz,0);
			file = file->next;
			count ++;
		}

		ZIP_CENTDIREND *hdrcdend = new ZIP_CENTDIREND;
		hdrcdend->sign			= ZIP_CENTDIREND_SIGN;
		hdrcdend->disknum		= 0;
		hdrcdend->diskcdstart	= 0;
		hdrcdend->countcddisk	= count;
		hdrcdend->countcd		= count;
		hdrcdend->cdsize		= SetFilePointer(ctx->zip_file,0,0,FILE_CURRENT) - off1;
		hdrcdend->cdoffset		= off1;
		hdrcdend->lencomm		= 0;
		WriteFile(ctx->zip_file,hdrcdend,sizeof(ZIP_CENTDIREND),&sz,0);
	}
	CloseHandle(ctx->zip_file);
	return true;
}

ZipArchiveEntry *ZipArchive::getEntry(std::string path) {
	for( auto iter = ctx->files.begin(); iter != ctx->files.end(); iter++ ) {
		ZipArchiveEntry *entry = *iter;
		if( entry->path == path )
			return entry;
	}
	return 0;
}

bool ZipArchive::writeFile(const char *name, const void *data, long size) {
	ZIP_LOCALFILEHDR *hdrlocal = new ZIP_LOCALFILEHDR;
	ZIP_CENTDIRFILE  *hdrcd    = new ZIP_CENTDIRFILE;

	hdrlocal->sign			= ZIP_LOCALFILEHDR_SIGN;
	hdrlocal->ver			= ZIP_VER;
	hdrlocal->flags			= 0;
	hdrlocal->compr			= 0;
	hdrlocal->lastmodtime	= 0;
	hdrlocal->lastmoddate	= 0;
	hdrlocal->crc32			= crc32(0,(uint8_t*)data,size);
	hdrlocal->sizecompr		= size;
	hdrlocal->sizeorig		= size;
	hdrlocal->lenname		= (uint16_t) strlen(name);
	hdrlocal->lenextr		= 0;

	hdrcd->sign			= ZIP_CENTDIRFILE_SIGN;
	hdrcd->ver			= ZIP_VER;
	hdrcd->vermade		= ZIP_VER;
	hdrcd->flags		= 0;
	hdrcd->compr		= 0;
	hdrcd->lastmodtime	= 0;
	hdrcd->lastmoddate	= 0;
	hdrcd->crc32		= hdrlocal->crc32;
	hdrcd->sizecompr	= size;
	hdrcd->sizeorig		= size;
	hdrcd->lenname		= (uint16_t) strlen(name);
	hdrcd->lenextr		= 0;
	hdrcd->lencomm		= 0;
	hdrcd->disknum		= 0;
	hdrcd->intattr		= 0;
	hdrcd->extattr		= 0;
	hdrcd->offset		= SetFilePointer(ctx->zip_file,0,0,FILE_CURRENT);

	DWORD sz;
	WriteFile(ctx->zip_file,hdrlocal,sizeof(ZIP_LOCALFILEHDR),&sz,0);
	WriteFile(ctx->zip_file,name,(uint32_t)strlen(name),&sz,0);
	WriteFile(ctx->zip_file,data,size,&sz,0);

	ZIP_FILE *file = new ZIP_FILE;
	file->d = hdrcd;
	file->next = 0;
	file->name = new char[strlen(name)+1];
	strcpy(file->name,name);
	if( ctx->file_last==0 ){
		ctx->file_first = file;
		ctx->file_last  = file;
	} else {
		ctx->file_last->next = file;
		ctx->file_last = file;
	}

	delete hdrlocal;
	return true;
}

bool ZipArchive::addFile(const char *extname, const char *intname) {
	HANDLE file = CreateFileA(extname,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,0,0);
	if( file==INVALID_HANDLE_VALUE )
		return false;
	unsigned long size = GetFileSize(file,0);
	void *data = VirtualAlloc(0,size,MEM_COMMIT,PAGE_READWRITE);
	DWORD sz;
	ReadFile(file,data,size,&sz,0);
	CloseHandle(file);
	writeFile(intname,data,sz);
	VirtualFree(data,size,MEM_RELEASE);
	return true;
}

ZipArchiveEntry::ZipArchiveEntry(ZipArchive *_zip, ZipArchiveEntry_Ctx *_ctx) {
	ctx = _ctx;
	zip = _zip;
	path   = ctx->path;
	size   = ctx->hdr_cntdirfile.sizeorig;
	opened = false;
}

ZipArchiveEntry::~ZipArchiveEntry() {
	delete ctx;
}

int ZipArchiveEntry::open(bool writable) {
	// TODO: writable
	DWORD sz;
	opened = true;
	SetFilePointer(zip->ctx->zip_file,ctx->hdr_cntdirfile.offset,0,FILE_BEGIN);
	ReadFile(zip->ctx->zip_file,&ctx->hdr_localfile,sizeof(ZIP_LOCALFILEHDR),&sz,0);
	if( ctx->hdr_localfile.sign != ZIP_LOCALFILEHDR_SIGN )
		return 1;
	ctx->offset = ctx->hdr_cntdirfile.offset + sizeof(ZIP_LOCALFILEHDR)
		+ ctx->hdr_localfile.lenname + ctx->hdr_localfile.lenextr;
	offset = 0;
	ctx->crc32 = 0;
	if( ctx->hdr_localfile.compr == ZIP_DEFLATE ) {
		ctx->buf0 = new unsigned char[ZCOMPR_BUFSIZE];
		ctx->buf1 = new unsigned char[ZCOMPR_BUFSIZE];
		memset(&ctx->z_str, 0, sizeof(ctx->z_str));
		ctx->z_str.next_in   = ctx->buf0;
		ctx->z_str.avail_in  = 0;
		ctx->z_str.next_out  = ctx->buf1;
		ctx->z_str.avail_out = ZCOMPR_BUFSIZE;
		if( inflateInit2(&ctx->z_str,-MAX_WBITS) )
			return 1;
		/*ctx->buf0 = (unsigned char*) VirtualAlloc(0,ctx->hdr_localfile.sizecompr,MEM_COMMIT,PAGE_READWRITE);
		ctx->buf1 = (unsigned char*) VirtualAlloc(0,ctx->hdr_localfile.sizeorig ,MEM_COMMIT,PAGE_READWRITE);
		SetFilePointer(zip->ctx->zip_file,ctx->offset,0,FILE_BEGIN);
		ReadFile(zip->ctx->zip_file,ctx->buf0,(DWORD)ctx->hdr_localfile.sizecompr,&sz,0);
		DWORD sz1 = ctx->hdr_localfile.sizeorig+200;
		int status = mz_compress2(ctx->buf1,&sz1,ctx->buf0,ctx->hdr_localfile.sizecompr,6);
		//sz = ctx->hdr_localfile.sizeorig;*
		//int status = mz_uncompress(ctx->buf1,&sz,ctx->buf0,ctx->hdr_localfile.sizecompr);
		//status = 0;
	}
	return 0;
}

int ZipArchiveEntry::read(void *data, size_t size, size_t *readed) {
	DWORD sz;
	if( ctx->hdr_localfile.compr == ZIP_DEFLATE ) {
		for(;;) {
			if( ctx->z_str.avail_in == 0) {
				uint32_t toread = ctx->hdr_localfile.sizecompr - offset;
				if( toread == 0 )
					break;
				if( toread > ZCOMPR_BUFSIZE )
					toread = ZCOMPR_BUFSIZE;
				SetFilePointer(zip->ctx->zip_file,ctx->offset+offset,0,FILE_BEGIN);
				ReadFile(zip->ctx->zip_file,ctx->buf0,toread,&sz,0);
				offset += sz;
				ctx->z_str.avail_in = toread;
				ctx->z_str.next_in  = ctx->buf0;
			}
			uint8_t *data = ctx->z_str.next_out;
			uint32_t size = ctx->z_str.total_out;
			int status = inflate(&ctx->z_str,Z_SYNC_FLUSH);
			if( status!=Z_OK )
				return 1;
			ctx->crc32 = crc32(ctx->crc32,data,ctx->z_str.total_out-size);
		}
	} else {
		SetFilePointer(zip->ctx->zip_file,ctx->offset+offset,0,FILE_BEGIN);
		ReadFile(zip->ctx->zip_file,data,(DWORD)size,&sz,0);
		offset += sz;
		*readed = sz;
	}
	return 0;
}

int ZipArchiveEntry::write(const void *data, size_t size, size_t *written) {
	return 0;
}
*/
