#define _CRT_SECURE_NO_WARNINGS
//#define _CSR_INTERNAL_

//#include <Windows.h>
//#include <iostream>
#include <string>
#include <vector>

#include "csrzip.h"
#include "csrzip_int.h"

Time time_from_dos(uint32_t dos);

FileBaseZip::FileBaseZip(ZipArchive *_zip, String _path) {
	zip = _zip;
	b_path = _path;
	b_file = NULL;
	f_dir  = false;
	Url url = b_path;
	if( String(url)==L"" )
		f_dir = true;
	else for( ZIP_FILE* file: zip->ctx->files ) {
		if( file->url == url ) {
			b_file = file;
			f_dir = file->url.format(URL_DIRECTORY);
			break;
		}
		if( file->url.match_as_parent(url) ) {
			f_dir = true;
			break;
		}
	}
}

bool FileBaseZip::getexists() {
	return b_file != NULL;
}

FileType FileBaseZip::gettype() {
	if( f_dir )
		return FileType_Directory;
	if( b_file != NULL )
		return FileType_File;
	else
		return FileType_Null;
}

uint64_t FileBaseZip::getsize() {
	if( b_file==NULL )
		return 0;
	return b_file->hdr_cntdirfile.sizeorig;
}

FileMD5 FileBaseZip::md5() {
	// TODO:
	return FileMD5();
}

FileHandle* FileBaseZip::open(uint32_t flags) {
	return new ZipArchiveEntryHandle(File(this),flags);
}

bool FileBaseZip::remove() {
	// TODO:
	return false;
}

bool FileBaseZip::createdir() {
	// TODO:
	return true;
}

FileIteratorBase* FileBaseZip::getfiles() {
	return new FileIteratorBaseZip(this);
}

FileBase* FileBaseZip::getchild(String path) {
	Url url = Url(b_path) + path;
	return new FileBaseZip(zip,url);
}

StringB FileBaseZip::getname() {
	return Url(b_path).getname();
}

void FileBaseZip::info(FILE_INFO& i) {
	i.fiSize = b_file->hdr_cntdirfile.sizeorig;
	Time time = time_from_dos(
		((uint32_t)b_file->hdr_cntdirfile.lastmoddate<<16)
		| b_file->hdr_cntdirfile.lastmodtime );
	i.flAccessed = time;
	i.flModified = time;
	i.flCreated  = time;
}

Url FileBaseZip::getpath() {
	return b_path;
}

FileBase* FileBaseZip::getfile(const Url& path) {
	return new FileBaseZip(zip,path);
}

FileBase* FileBaseZip::clone() {
	return new FileBaseZip(zip,b_path);
}

FileIteratorBaseZip::FileIteratorBaseZip(FileBase* directory) {
	dir = (FileBaseZip*) directory;
	zip = dir->zip;
	Url base = Url(dir->b_path).p_file();
	for( ZIP_FILE* file: zip->ctx->files ) {
		if( base.is_inside_imm(file->url) ) {
			list.push_back(file->path);
		}
	}
	list_pos = 0;
}

FileBase* FileIteratorBaseZip::getnext() {
	if( list_pos >= list.size() )
		return NULL;
	return new FileBaseZip(zip,list[list_pos++]);
}

void FileIteratorBaseZip::begin() {
	list_pos = 0;
}

Time time_from_dos(uint32_t dos) {
	TimeInfo ti;
	ti.millisecond = 0;
	ti.second      = 2 * (dos&0x1F);
	ti.minute      = (dos>>5)&0x3F;
	ti.hour        = (dos>>11)&0x1F;
	ti.day         = (dos>>16)&0x1F;
	ti.month       = (dos>>21)&0x0F;
	ti.year        = 1980 + ((dos>>25)&0x7F);
	return ti;
}

