#pragma once

#ifndef _CSR_INTERNAL_

#error "Unnecessary inclusion of internal header."

#else

//unsigned long crc32(unsigned long crc, const void *buf, size_t size);

#include <zlib.h>
//#include <Windows.h>

typedef unsigned int DWORD;
typedef unsigned short WORD;

using namespace csrutils;

#define ZIP_VER			0x0A
#define ZIP_DEFLATE		0x0008

#define ZCOMPR_BUFSIZE	1024

#pragma pack(push)  /* push current alignment to stack */
#pragma pack(1)

#define ZIP_LOCALFILEHDR_SIGN 0x04034b50
struct ZIP_LOCALFILEHDR {
	DWORD sign;			// local file header signature     4 bytes  (0x04034b50)
	WORD ver;			// version needed to extract       2 bytes
	WORD flags;			// general purpose bit flag        2 bytes
	WORD compr;			// compression method              2 bytes
	WORD lastmodtime;	// last mod file time              2 bytes
	WORD lastmoddate;	// last mod file date              2 bytes
	DWORD crc32;		// crc-32                          4 bytes
	DWORD sizecompr;	// compressed size                 4 bytes
	DWORD sizeorig;		// uncompressed size               4 bytes
	WORD lenname;		// file name length                2 bytes
	WORD lenextr;		// extra field length              2 bytes
};

#define ZIP_CENTDIRFILE_SIGN 0x02014b50
struct ZIP_CENTDIRFILE {
	DWORD sign;			// central file header signature   4 bytes  (0x02014b50)
	WORD vermade;		// version made by                 2 bytes
	WORD ver;			// version needed to extract       2 bytes
	WORD flags;			// general purpose bit flag        2 bytes
	WORD compr;			// compression method              2 bytes
	WORD lastmodtime;	// last mod file time              2 bytes
	WORD lastmoddate;	// last mod file date              2 bytes
	DWORD crc32;		// crc-32                          4 bytes
	DWORD sizecompr;	// compressed size                 4 bytes
	DWORD sizeorig;		// uncompressed size               4 bytes
	WORD lenname;		// file name length                2 bytes
	WORD lenextr;		// extra field length              2 bytes
	WORD lencomm;		// file comment length             2 bytes
	WORD disknum;		// disk number start               2 bytes
	WORD intattr;		// internal file attributes        2 bytes
	DWORD extattr;		// external file attributes        4 bytes
	DWORD offset;		// relative offset of local header 4 bytes
};

#define ZIP_CENTDIREND_SIGN 0x06054b50
struct ZIP_CENTDIREND {
	DWORD sign;			// end of central dir signature    4 bytes  (0x06054b50)
	WORD disknum;		// number of this disk             2 bytes
	WORD diskcdstart;	// number of the disk with the
						// start of the central directory  2 bytes
	WORD countcddisk;	// total number of entries in the
						// central directory on this disk  2 bytes
	WORD countcd;		// total number of entries in
						// the central directory           2 bytes
	DWORD cdsize;		// size of the central directory   4 bytes
	DWORD cdoffset;		// offset of start of central
						// directory with respect to
						// the starting disk number        4 bytes
	WORD lencomm;		// .ZIP file comment length        2 bytes
};

#pragma pack(pop)

struct ZIP_FILE {
	ZIP_CENTDIRFILE hdr_cntdirfile;
	String path;
	void *extra;
	String comment;
	Url url;
};

struct csrutils::ZipArchive_Ctx {
	//HANDLE zip_file;
	File *file;
	FileHandle *file_handle;
	//ZIP_FILE *file_first, *file_last;
	ZipArchiveMode zip_mode;
	ZIP_CENTDIREND hdr_cntdir;
	std::vector<ZIP_FILE*> files;
	size_t end_of_data;
};

struct csrutils::ZipArchiveEntryHandle_Ctx {
	ZIP_CENTDIRFILE hdr_cntdirfile;
	ZIP_LOCALFILEHDR hdr_localfile;
	unsigned long offset;		// offset of data
	std::string path;
	z_stream_s z_str;
	unsigned char *buf0, *buf1;
	size_t buf1_off;
	unsigned long crc32;
	uint32_t open_flags;
	ZIP_FILE* zip_file;
};

class FileBaseZip;
class FileIteratorBaseZip;

class FileBaseZip : public csrutils::FileBase {
	friend class csrutils::ZipArchiveEntryHandle;

	String b_path;
	ZipArchive *zip;
	ZIP_FILE* b_file;
	bool f_dir;

public:
	FileBaseZip(ZipArchive *zip, String path);
	//FileBaseZip(const FileBaseZip& base);
	~FileBaseZip() { };

	bool getexists();
	FileType gettype();
	uint64_t getsize();
	FileMD5 md5();

	FileHandle* open(uint32_t flags);
	FileIteratorBase* getfiles();
	FileBase* getchild(String path);
	StringB getname();
	void info(FILE_INFO&);
	Url getpath();
	FileBase* getfile(const Url& path);
	
	bool remove();
	bool createdir();

	FileBase* clone();

	inline String getpath() const { return b_path; }

	friend class FileIteratorBaseZip;
};

class FileIteratorBaseZip : public FileIteratorBase {
	FileBaseZip* dir;
	bool found;
	ZipArchive* zip;
	std::vector<String> list;
	size_t list_pos;

public:
	FileIteratorBaseZip(FileBase* directory);
	~FileIteratorBaseZip();

	FileBase* getnext();
	void begin();

	friend class FileBaseZip;
};

/*extern "C" {

typedef unsigned long mz_ulong;

typedef void *(*mz_alloc_func)(void *opaque, size_t items, size_t size);
typedef void (*mz_free_func)(void *opaque, void *address);
typedef void *(*mz_realloc_func)(void *opaque, void *address, size_t items, size_t size);

int mz_compress(unsigned char *pDest, mz_ulong *pDest_len, const unsigned char *pSource, mz_ulong source_len);
int mz_compress2(unsigned char *pDest, mz_ulong *pDest_len, const unsigned char *pSource, mz_ulong source_len, int level);
int mz_uncompress(unsigned char *pDest, mz_ulong *pDest_len, const unsigned char *pSource, mz_ulong source_len);

typedef struct mz_stream_s
{
  const unsigned char *next_in;     // pointer to next byte to read
  unsigned int avail_in;            // number of bytes available at next_in
  mz_ulong total_in;                // total number of bytes consumed so far

  unsigned char *next_out;          // pointer to next byte to write
  unsigned int avail_out;           // number of bytes that can be written to next_out
  mz_ulong total_out;               // total number of bytes produced so far

  char *msg;                        // error msg (unused)
  struct mz_internal_state *state;  // internal state, allocated by zalloc/zfree

  mz_alloc_func zalloc;             // optional heap allocation function (defaults to malloc)
  mz_free_func zfree;               // optional heap free function (defaults to free)
  void *opaque;                     // heap alloc function user pointer

  int data_type;                    // data_type (unused)
  mz_ulong adler;                   // adler32 of the source or uncompressed data
  mz_ulong reserved;                // not used
} mz_stream;

typedef mz_stream *mz_streamp;

enum { MZ_NO_FLUSH = 0, MZ_PARTIAL_FLUSH = 1, MZ_SYNC_FLUSH = 2, MZ_FULL_FLUSH = 3, MZ_FINISH = 4, MZ_BLOCK = 5 };
enum { MZ_OK = 0, MZ_STREAM_END = 1, MZ_NEED_DICT = 2, MZ_ERRNO = -1, MZ_STREAM_ERROR = -2, MZ_DATA_ERROR = -3, MZ_MEM_ERROR = -4, MZ_BUF_ERROR = -5, MZ_VERSION_ERROR = -6, MZ_PARAM_ERROR = -10000 };

int mz_inflateInit(mz_streamp pStream);
int mz_inflate(mz_streamp pStream, int flush);
int mz_inflateEnd(mz_streamp pStream);

}*/

#endif
