#define _CRT_SECURE_NO_WARNINGS
//#define _CSR_INTERNAL_

//#include <Windows.h>
//#include <iostream>
#include <string>
#include <vector>

#include "csrzip.h"
#include "csrzip_int.h"

ZipArchiveEntryHandle::ZipArchiveEntryHandle(const File& file, uint32_t flags) {
	ctx = new ZipArchiveEntryHandle_Ctx;
	size_t sz;
	FileBaseZip* fbase = (FileBaseZip*) file.getbase();
	zip = fbase->zip;
	ctx->open_flags = flags;
	ctx->crc32 = 0;
	ctx->buf0 = NULL;
	ctx->buf1 = NULL;
	offset = 0;
	if( (flags&File_Read) && !(flags&File_Write) ) {
		if( !fbase->getexists() )
			throw 0;
		if( fbase->gettype() != FileType_File )
			throw 0;
		ctx->hdr_cntdirfile = fbase->b_file->hdr_cntdirfile;
		//ctx->offset = 0;
		//SetFilePointer(zip->ctx->zip_file,ctx->hdr_cntdirfile.offset,0,FILE_BEGIN);
		//ReadFile(zip->ctx->zip_file,&ctx->hdr_localfile,sizeof(ZIP_LOCALFILEHDR),&sz,0);
		zip->ctx->file_handle->setposition(ctx->hdr_cntdirfile.offset);
		zip->ctx->file_handle->read(&ctx->hdr_localfile,sizeof(ZIP_LOCALFILEHDR),sz);
		if( ctx->hdr_localfile.sign != ZIP_LOCALFILEHDR_SIGN )
			throw 1;
		ctx->offset = ctx->hdr_cntdirfile.offset + sizeof(ZIP_LOCALFILEHDR)
			+ ctx->hdr_localfile.lenname + ctx->hdr_localfile.lenextr;
		if( ctx->hdr_localfile.compr == ZIP_DEFLATE ) {
			ctx->buf0 = new unsigned char[ZCOMPR_BUFSIZE];
			ctx->buf1 = new unsigned char[ZCOMPR_BUFSIZE];
			ctx->buf1_off = 0;
			memset(&ctx->z_str, 0, sizeof(ctx->z_str));
			ctx->z_str.next_in   = ctx->buf0;
			ctx->z_str.avail_in  = 0;
			ctx->z_str.next_out  = ctx->buf1;
			ctx->z_str.avail_out = ZCOMPR_BUFSIZE;
			if( inflateInit2(&ctx->z_str,-MAX_WBITS) )
				throw 1;
		}
	} else if(flags&File_Write) {
		String name = file.path();
		ZIP_LOCALFILEHDR *hdrlocal = &ctx->hdr_localfile;
		ZIP_CENTDIRFILE  *hdrcd    = &ctx->hdr_cntdirfile;
		hdrlocal->sign			= ZIP_LOCALFILEHDR_SIGN;
		hdrlocal->ver			= ZIP_VER;
		hdrlocal->flags			= 0;
		hdrlocal->compr			= 0;
		hdrlocal->lastmodtime	= 0;
		hdrlocal->lastmoddate	= 0;
		hdrlocal->crc32			= 0;
		hdrlocal->sizecompr		= 0;
		hdrlocal->sizeorig		= 0;
		hdrlocal->lenname		= (uint16_t) name.len();
		hdrlocal->lenextr		= 0;
		hdrcd->sign			= ZIP_CENTDIRFILE_SIGN;
		hdrcd->ver			= ZIP_VER;
		hdrcd->vermade		= ZIP_VER;
		hdrcd->flags		= 0;
		hdrcd->compr		= 0;
		hdrcd->lastmodtime	= 0;
		hdrcd->lastmoddate	= 0;
		hdrcd->crc32		= hdrlocal->crc32;
		hdrcd->sizecompr	= 0;
		hdrcd->sizeorig		= 0;
		hdrcd->lenname		= (uint16_t) name.len();
		hdrcd->lenextr		= 0;
		hdrcd->lencomm		= 0;
		hdrcd->disknum		= 0;
		hdrcd->intattr		= 0;
		hdrcd->extattr		= 0;
		hdrcd->offset		= (DWORD) zip->ctx->end_of_data;
		zip->ctx->file_handle->setposition( zip->ctx->end_of_data );
		size_t sz;
		zip->ctx->file_handle->write(hdrlocal,sizeof(ZIP_LOCALFILEHDR),sz);
		zip->ctx->file_handle->write(name.utf8().bytes(),name.len(),sz);
		if( !fbase->getexists() ) {
			ctx->zip_file = new ZIP_FILE;
			ctx->zip_file->extra = new char[0];
			ctx->zip_file->path  = name;
			ctx->zip_file->url   = name;
			ctx->zip_file->hdr_cntdirfile = ctx->hdr_cntdirfile;
			zip->ctx->files.push_back(ctx->zip_file);
		} else if( fbase->gettype() != FileType_File )
			throw 0;
		else {
			ctx->zip_file = fbase->b_file;
			ctx->zip_file->extra = new char[0];
			ctx->zip_file->path  = name;
			ctx->zip_file->url   = name;
			ctx->zip_file->hdr_cntdirfile = ctx->hdr_cntdirfile;
		}

	}
}

ZipArchiveEntryHandle::~ZipArchiveEntryHandle() {
	if( ctx->open_flags&File_Write ) {
		size_t sz;
		zip->ctx->end_of_data = zip->ctx->file_handle->getposition();
		ctx->hdr_cntdirfile.sizecompr = (DWORD) offset;
		ctx->hdr_cntdirfile.sizeorig  = (DWORD) offset;
		ctx->hdr_cntdirfile.crc32     = ctx->crc32;
		ctx->hdr_localfile.sizecompr  = (DWORD) offset;
		ctx->hdr_localfile.sizeorig   = (DWORD) offset;
		ctx->hdr_localfile.crc32      = ctx->crc32;
		ctx->zip_file->hdr_cntdirfile = ctx->hdr_cntdirfile;
		zip->ctx->file_handle->setposition( ctx->hdr_cntdirfile.offset );
		zip->ctx->file_handle->write(&ctx->hdr_localfile,sizeof(ZIP_LOCALFILEHDR),sz);
	}
	delete ctx;
}

int ZipArchiveEntryHandle::read(void *data, size_t size, size_t& read) {
	char *_data = (char*) data;
	size_t sz;
	bool s_end = false;
	if( ctx->hdr_localfile.compr == ZIP_DEFLATE ) {
		size_t sz_rd = 0;
		while( sz_rd < size && (!s_end) ) {
			if( ctx->buf1_off == ctx->z_str.total_out ) {
				if( ctx->z_str.avail_in == 0) {
					ctx->z_str.next_in = ctx->buf0;
					size_t toread = ctx->hdr_localfile.sizecompr - offset;
					if( toread == 0 )
						s_end = true;
					if( toread > ZCOMPR_BUFSIZE )
						toread = ZCOMPR_BUFSIZE;
					//SetFilePointer(zip->ctx->zip_file,ctx->offset+offset,0,FILE_BEGIN);
					//ReadFile(zip->ctx->zip_file,ctx->buf0,toread,&sz,0);
					zip->ctx->file_handle->setposition(ctx->offset+offset);
					zip->ctx->file_handle->read(ctx->buf0,toread,sz);
					offset += sz;
					ctx->z_str.avail_in = (int) sz;
					ctx->z_str.next_in  = ctx->buf0;
				}
				ctx->z_str.next_out  = ctx->buf1;
				ctx->z_str.total_out = 0;
				ctx->z_str.avail_out = ZCOMPR_BUFSIZE;
				ctx->buf1_off        = 0;
				int status = inflate(&ctx->z_str,Z_SYNC_FLUSH);
				if( !( status==Z_OK || status==Z_STREAM_END ) )
					return 1;
				if( status==Z_STREAM_END )
					s_end = true;
				ctx->crc32 = crc32(ctx->crc32,ctx->buf1,ctx->z_str.total_out);
			}
			size_t sz1 = size - sz_rd;
			if( sz1 > ctx->z_str.total_out-ctx->buf1_off )
				sz1 = ctx->z_str.total_out - ctx->buf1_off;
			memcpy( _data+sz_rd, ctx->buf1+ctx->buf1_off, sz1 );
			sz_rd += sz1;
			ctx->buf1_off += sz1;
		}
		read = sz_rd;
	} else {
		//SetFilePointer(zip->ctx->zip_file,ctx->offset+offset,0,FILE_BEGIN);
		//ReadFile(zip->ctx->zip_file,data,(DWORD)size,&sz,0);
		zip->ctx->file_handle->setposition(ctx->offset+offset);
		zip->ctx->file_handle->read(ctx->buf0,size,sz);
		offset += sz;
		read = sz;
	}
	return 0;
}

int ZipArchiveEntryHandle::write(const void *data, size_t size, size_t& written) {
	size_t sz = 0;
	zip->ctx->file_handle->write(data,size,sz);
	ctx->crc32 = crc32(ctx->crc32,(uint8_t*)data,(int)sz);
	offset += sz;
	written = sz;
	return 0;
}

uint64_t ZipArchiveEntryHandle::getsize() {
	return ctx->hdr_localfile.sizeorig;
}

uint64_t ZipArchiveEntryHandle::getposition() {
	if( ctx->hdr_localfile.compr == 0 )
		return offset;
	return 0;
}

bool ZipArchiveEntryHandle::setposition(uint64_t pos) {
	if( ctx->hdr_localfile.compr == 0 )
		return zip->ctx->file_handle->setposition( ctx->offset + offset );
	return false;
}
