#define _CRT_SECURE_NO_WARNINGS

#include <Windows.h>
#include <iostream>
#include <string>

#include "zip.h"

#define ZIP_VER		0x0A

#pragma pack(push)  /* push current alignment to stack */
#pragma pack(1)

#define ZIP_LOCALFILEHDR_SIGN 0x04034b50
struct ZIP_LOCALFILEHDR {
	DWORD sign;			// local file header signature     4 bytes  (0x04034b50)
	WORD ver;			// version needed to extract       2 bytes
	WORD flags;			// general purpose bit flag        2 bytes
	WORD compr;			// compression method              2 bytes
	WORD lastmodtime;	// last mod file time              2 bytes
	WORD lastmoddate;	// last mod file date              2 bytes
	DWORD crc32;		// crc-32                          4 bytes
	DWORD sizecompr;	// compressed size                 4 bytes
	DWORD sizeorig;		// uncompressed size               4 bytes
	WORD lenname;		// file name length                2 bytes
	WORD lenextr;		// extra field length              2 bytes
};

#define ZIP_CENTDIRFILE_SIGN 0x02014b50
struct ZIP_CENTDIRFILE {
	DWORD sign;			// central file header signature   4 bytes  (0x02014b50)
	WORD vermade;		// version made by                 2 bytes
	WORD ver;			// version needed to extract       2 bytes
	WORD flags;			// general purpose bit flag        2 bytes
	WORD compr;			// compression method              2 bytes
	WORD lastmodtime;	// last mod file time              2 bytes
	WORD lastmoddate;	// last mod file date              2 bytes
	DWORD crc32;		// crc-32                          4 bytes
	DWORD sizecompr;	// compressed size                 4 bytes
	DWORD sizeorig;		// uncompressed size               4 bytes
	WORD lenname;		// file name length                2 bytes
	WORD lenextr;		// extra field length              2 bytes
	WORD lencomm;		// file comment length             2 bytes
	WORD disknum;		// disk number start               2 bytes
	WORD intattr;		// internal file attributes        2 bytes
	DWORD extattr;		// external file attributes        4 bytes
	DWORD offset;		// relative offset of local header 4 bytes
};

#define ZIP_CENTDIREND_SIGN 0x06054b50
struct ZIP_CENTDIREND {
	DWORD sign;			// end of central dir signature    4 bytes  (0x06054b50)
	WORD disknum;		// number of this disk             2 bytes
	WORD diskcdstart;	// number of the disk with the
						// start of the central directory  2 bytes
	WORD countcddisk;	// total number of entries in the
						// central directory on this disk  2 bytes
	WORD countcd;		// total number of entries in
						// the central directory           2 bytes
	DWORD cdsize;		// size of the central directory   4 bytes
	DWORD cdoffset;		// offset of start of central
						// directory with respect to
						// the starting disk number        4 bytes
	WORD lencomm;		// .ZIP file comment length        2 bytes
};

#pragma pack(pop)

struct ZIP_FILE {
	ZIP_CENTDIRFILE *d;
	ZIP_FILE *next;
	char *name;
};

unsigned long crc32(unsigned long crc, const void *buf, size_t size);

bool ZipArchive::open(const char *file) {
	zip_file   = CreateFileA(file,GENERIC_WRITE,0,0,CREATE_ALWAYS,0,0);
	file_first = 0;
	file_last  = 0;
	return true;
}

bool ZipArchive::close() {
	DWORD sz;

	DWORD off1 = SetFilePointer(zip_file,0,0,FILE_CURRENT);

	ZIP_FILE *file = file_first;
	WORD count = 0;
	while(file) {
		WriteFile(zip_file,file->d,sizeof(ZIP_CENTDIRFILE),&sz,0);
		WriteFile(zip_file,file->name,strlen(file->name),&sz,0);
		file = file->next;
		count ++;
	}

	ZIP_CENTDIREND *hdrcdend = new ZIP_CENTDIREND;
	hdrcdend->sign			= ZIP_CENTDIREND_SIGN;
	hdrcdend->disknum		= 0;
	hdrcdend->diskcdstart	= 0;
	hdrcdend->countcddisk	= count;
	hdrcdend->countcd		= count;
	hdrcdend->cdsize		= SetFilePointer(zip_file,0,0,FILE_CURRENT) - off1;
	hdrcdend->cdoffset		= off1;
	hdrcdend->lencomm		= 0;
	WriteFile(zip_file,hdrcdend,sizeof(ZIP_CENTDIREND),&sz,0);

	CloseHandle(zip_file);
	return true;
}

bool ZipArchive::writeFile(const char *name, const void *data, long size) {
	ZIP_LOCALFILEHDR *hdrlocal = new ZIP_LOCALFILEHDR;
	ZIP_CENTDIRFILE  *hdrcd    = new ZIP_CENTDIRFILE;

	hdrlocal->sign			= ZIP_LOCALFILEHDR_SIGN;
	hdrlocal->ver			= ZIP_VER;
	hdrlocal->flags			= 0;
	hdrlocal->compr			= 0;
	hdrlocal->lastmodtime	= 0;
	hdrlocal->lastmoddate	= 0;
	hdrlocal->crc32			= crc32(0,data,size);
	hdrlocal->sizecompr		= size;
	hdrlocal->sizeorig		= size;
	hdrlocal->lenname		= strlen(name);
	hdrlocal->lenextr		= 0;

	hdrcd->sign			= ZIP_CENTDIRFILE_SIGN;
	hdrcd->ver			= ZIP_VER;
	hdrcd->vermade		= ZIP_VER;
	hdrcd->flags		= 0;
	hdrcd->compr		= 0;
	hdrcd->lastmodtime	= 0;
	hdrcd->lastmoddate	= 0;
	hdrcd->crc32		= hdrlocal->crc32;
	hdrcd->sizecompr	= size;
	hdrcd->sizeorig		= size;
	hdrcd->lenname		= strlen(name);
	hdrcd->lenextr		= 0;
	hdrcd->lencomm		= 0;
	hdrcd->disknum		= 0;
	hdrcd->intattr		= 0;
	hdrcd->extattr		= 0;
	hdrcd->offset		= SetFilePointer(zip_file,0,0,FILE_CURRENT);

	DWORD sz;
	WriteFile(zip_file,hdrlocal,sizeof(ZIP_LOCALFILEHDR),&sz,0);
	WriteFile(zip_file,name,strlen(name),&sz,0);
	WriteFile(zip_file,data,size,&sz,0);

	ZIP_FILE *file = new ZIP_FILE;
	file->d = hdrcd;
	file->next = 0;
	file->name = new char[strlen(name)+1];
	strcpy(file->name,name);
	if( file_last==0 ){
		file_first = file;
		file_last  = file;
	} else {
		file_last->next = file;
		file_last = file;
	}

	delete hdrlocal;
	return true;
}

bool ZipArchive::addFile(const char *extname, const char *intname) {
	HANDLE file = CreateFileA(extname,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,0,0);
	if( file==INVALID_HANDLE_VALUE )
		return false;
	unsigned long size = GetFileSize(file,0);
	void *data = VirtualAlloc(0,size,MEM_COMMIT,PAGE_READWRITE);
	DWORD sz;
	ReadFile(file,data,size,&sz,0);
	CloseHandle(file);
	writeFile(intname,data,sz);
	VirtualFree(data,size,MEM_RELEASE);
	return true;
}

