#include <csrutils.h>
#include <csrmem.h>
#include <csrstr.h>
#include <csrheap.h>

using namespace csrutils;

struct csrutils::CsrMemoryContextUser {
};

THREAD_LOCAL CsrMemoryContext csrmem_thread_context;

void *csrutils::heap_malloc(size_t size, uint64_t flags) {
	return csrmem_thread_context.alloc(size,flags);
}

void csrutils::heap_mfree(void* data, uint64_t flags) {
	csrmem_thread_context.free(data,flags);
}

void* csr_cdefault_alloc(size_t sz, uint64_t flags) {
	return ::malloc(sz);
}

void csr_cdefault_free(void* ref, uint64_t flags) {
	::free(ref);
}

void* csr_cppdefault_alloc(size_t sz, uint64_t flags) {
	return ::new char[sz];
}

void csr_cppdefault_free(void* ref, uint64_t flags) {
	::delete[] (char*) ref;
}

void* csr_heap_alloc(size_t sz, uint64_t flags) {
	return heap_alloc( (Heap_Ctx*) csrmem_thread_context.ctx, sz );
}

void csr_heap_free(void* ref, uint64_t flags) {
	heap_free( (Heap_Ctx*) csrmem_thread_context.ctx, ref );
}

void csrutils::heap_set_cdefault() {
	csrmem_thread_context.ctx   = nullptr;
	csrmem_thread_context.alloc = csr_cdefault_alloc;
	csrmem_thread_context.free  = csr_cdefault_free;
	csrmem_thread_context.prev  = nullptr;
}

void csrutils::heap_set_cppdefault() {
	csrmem_thread_context.ctx   = nullptr;
	csrmem_thread_context.alloc = csr_cppdefault_alloc;
	csrmem_thread_context.free  = csr_cppdefault_free;
	csrmem_thread_context.prev  = nullptr;
}

CSRLIB void csrutils::heap_set(Heap& h) {
	csrmem_thread_context.ctx   = (CsrMemoryContextUser*) h.getctx();
	csrmem_thread_context.alloc = csr_heap_alloc;
	csrmem_thread_context.free  = csr_heap_free;
}

CSRLIB void heap_push_ctx(const CsrMemoryContext &ctx) {
	CsrMemoryContext *prev = new ( heap_alloc<CsrMemoryContext>() ) CsrMemoryContext(csrmem_thread_context);
	csrmem_thread_context = ctx;
	csrmem_thread_context.prev = prev;
}

CSRLIB void heap_push(Heap& h) {
	CsrMemoryContext *prev = new ( heap_alloc<CsrMemoryContext>() ) CsrMemoryContext(csrmem_thread_context);
	csrmem_thread_context.ctx   = (CsrMemoryContextUser*) h.getctx();
	csrmem_thread_context.alloc = csr_heap_alloc;
	csrmem_thread_context.free  = csr_heap_free;
	csrmem_thread_context.prev  = prev;
}

CSRLIB void heap_pop() {
	CsrMemoryContext *prev = csrmem_thread_context.prev;
	csrmem_thread_context = *prev;
	heap_free(prev);
}
